# This file was auto-generated by Fern from our API Definition.

from .action_block import ActionBlock
from .action_block_data_schema import ActionBlockDataSchema
from .action_block_parameters_item import (
    ActionBlockParametersItem,
    ActionBlockParametersItem_AwsSecret,
    ActionBlockParametersItem_BitwardenCreditCardData,
    ActionBlockParametersItem_BitwardenLoginCredential,
    ActionBlockParametersItem_BitwardenSensitiveInformation,
    ActionBlockParametersItem_Context,
    ActionBlockParametersItem_Credential,
    ActionBlockParametersItem_Output,
    ActionBlockParametersItem_Workflow,
)
from .aws_secret_parameter import AwsSecretParameter
from .bitwarden_credit_card_data_parameter import BitwardenCreditCardDataParameter
from .bitwarden_login_credential_parameter import BitwardenLoginCredentialParameter
from .bitwarden_sensitive_information_parameter import BitwardenSensitiveInformationParameter
from .browser_session_response import BrowserSessionResponse
from .code_block import CodeBlock
from .code_block_parameters_item import (
    CodeBlockParametersItem,
    CodeBlockParametersItem_AwsSecret,
    CodeBlockParametersItem_BitwardenCreditCardData,
    CodeBlockParametersItem_BitwardenLoginCredential,
    CodeBlockParametersItem_BitwardenSensitiveInformation,
    CodeBlockParametersItem_Context,
    CodeBlockParametersItem_Credential,
    CodeBlockParametersItem_Output,
    CodeBlockParametersItem_Workflow,
)
from .context_parameter import ContextParameter
from .credential_parameter import CredentialParameter
from .download_to_s3block import DownloadToS3Block
from .extraction_block import ExtractionBlock
from .extraction_block_data_schema import ExtractionBlockDataSchema
from .extraction_block_parameters_item import (
    ExtractionBlockParametersItem,
    ExtractionBlockParametersItem_AwsSecret,
    ExtractionBlockParametersItem_BitwardenCreditCardData,
    ExtractionBlockParametersItem_BitwardenLoginCredential,
    ExtractionBlockParametersItem_BitwardenSensitiveInformation,
    ExtractionBlockParametersItem_Context,
    ExtractionBlockParametersItem_Credential,
    ExtractionBlockParametersItem_Output,
    ExtractionBlockParametersItem_Workflow,
)
from .file_download_block import FileDownloadBlock
from .file_download_block_data_schema import FileDownloadBlockDataSchema
from .file_download_block_parameters_item import (
    FileDownloadBlockParametersItem,
    FileDownloadBlockParametersItem_AwsSecret,
    FileDownloadBlockParametersItem_BitwardenCreditCardData,
    FileDownloadBlockParametersItem_BitwardenLoginCredential,
    FileDownloadBlockParametersItem_BitwardenSensitiveInformation,
    FileDownloadBlockParametersItem_Context,
    FileDownloadBlockParametersItem_Credential,
    FileDownloadBlockParametersItem_Output,
    FileDownloadBlockParametersItem_Workflow,
)
from .file_info import FileInfo
from .file_parser_block import FileParserBlock
from .file_storage_type import FileStorageType
from .file_type import FileType
from .file_upload_block import FileUploadBlock
from .for_loop_block import ForLoopBlock
from .for_loop_block_loop_blocks_item import (
    ForLoopBlockLoopBlocksItem,
    ForLoopBlockLoopBlocksItem_Action,
    ForLoopBlockLoopBlocksItem_Code,
    ForLoopBlockLoopBlocksItem_DownloadToS3,
    ForLoopBlockLoopBlocksItem_Extraction,
    ForLoopBlockLoopBlocksItem_FileDownload,
    ForLoopBlockLoopBlocksItem_FileUpload,
    ForLoopBlockLoopBlocksItem_FileUrlParser,
    ForLoopBlockLoopBlocksItem_ForLoop,
    ForLoopBlockLoopBlocksItem_GotoUrl,
    ForLoopBlockLoopBlocksItem_Login,
    ForLoopBlockLoopBlocksItem_Navigation,
    ForLoopBlockLoopBlocksItem_PdfParser,
    ForLoopBlockLoopBlocksItem_SendEmail,
    ForLoopBlockLoopBlocksItem_Task,
    ForLoopBlockLoopBlocksItem_TaskV2,
    ForLoopBlockLoopBlocksItem_TextPrompt,
    ForLoopBlockLoopBlocksItem_UploadToS3,
    ForLoopBlockLoopBlocksItem_Validation,
    ForLoopBlockLoopBlocksItem_Wait,
)
from .for_loop_block_loop_over import (
    ForLoopBlockLoopOver,
    ForLoopBlockLoopOver_AwsSecret,
    ForLoopBlockLoopOver_BitwardenCreditCardData,
    ForLoopBlockLoopOver_BitwardenLoginCredential,
    ForLoopBlockLoopOver_BitwardenSensitiveInformation,
    ForLoopBlockLoopOver_Context,
    ForLoopBlockLoopOver_Credential,
    ForLoopBlockLoopOver_Output,
    ForLoopBlockLoopOver_Workflow,
)
from .http_validation_error import HttpValidationError
from .login_block import LoginBlock
from .login_block_data_schema import LoginBlockDataSchema
from .login_block_parameters_item import (
    LoginBlockParametersItem,
    LoginBlockParametersItem_AwsSecret,
    LoginBlockParametersItem_BitwardenCreditCardData,
    LoginBlockParametersItem_BitwardenLoginCredential,
    LoginBlockParametersItem_BitwardenSensitiveInformation,
    LoginBlockParametersItem_Context,
    LoginBlockParametersItem_Credential,
    LoginBlockParametersItem_Output,
    LoginBlockParametersItem_Workflow,
)
from .navigation_block import NavigationBlock
from .navigation_block_data_schema import NavigationBlockDataSchema
from .navigation_block_parameters_item import (
    NavigationBlockParametersItem,
    NavigationBlockParametersItem_AwsSecret,
    NavigationBlockParametersItem_BitwardenCreditCardData,
    NavigationBlockParametersItem_BitwardenLoginCredential,
    NavigationBlockParametersItem_BitwardenSensitiveInformation,
    NavigationBlockParametersItem_Context,
    NavigationBlockParametersItem_Credential,
    NavigationBlockParametersItem_Output,
    NavigationBlockParametersItem_Workflow,
)
from .output import Output
from .output_parameter import OutputParameter
from .pdf_parser_block import PdfParserBlock
from .proxy_location import ProxyLocation
from .run_engine import RunEngine
from .run_status import RunStatus
from .send_email_block import SendEmailBlock
from .source import (
    Source,
    Source_AwsSecret,
    Source_BitwardenCreditCardData,
    Source_BitwardenLoginCredential,
    Source_BitwardenSensitiveInformation,
    Source_Context,
    Source_Credential,
    Source_Output,
    Source_Workflow,
)
from .task_block import TaskBlock
from .task_block_data_schema import TaskBlockDataSchema
from .task_block_parameters_item import (
    TaskBlockParametersItem,
    TaskBlockParametersItem_AwsSecret,
    TaskBlockParametersItem_BitwardenCreditCardData,
    TaskBlockParametersItem_BitwardenLoginCredential,
    TaskBlockParametersItem_BitwardenSensitiveInformation,
    TaskBlockParametersItem_Context,
    TaskBlockParametersItem_Credential,
    TaskBlockParametersItem_Output,
    TaskBlockParametersItem_Workflow,
)
from .task_run_request import TaskRunRequest
from .task_run_request_data_extraction_schema import TaskRunRequestDataExtractionSchema
from .task_run_response import TaskRunResponse
from .task_v2block import TaskV2Block
from .text_prompt_block import TextPromptBlock
from .text_prompt_block_parameters_item import (
    TextPromptBlockParametersItem,
    TextPromptBlockParametersItem_AwsSecret,
    TextPromptBlockParametersItem_BitwardenCreditCardData,
    TextPromptBlockParametersItem_BitwardenLoginCredential,
    TextPromptBlockParametersItem_BitwardenSensitiveInformation,
    TextPromptBlockParametersItem_Context,
    TextPromptBlockParametersItem_Credential,
    TextPromptBlockParametersItem_Output,
    TextPromptBlockParametersItem_Workflow,
)
from .upload_to_s3block import UploadToS3Block
from .url_block import UrlBlock
from .url_block_data_schema import UrlBlockDataSchema
from .url_block_parameters_item import (
    UrlBlockParametersItem,
    UrlBlockParametersItem_AwsSecret,
    UrlBlockParametersItem_BitwardenCreditCardData,
    UrlBlockParametersItem_BitwardenLoginCredential,
    UrlBlockParametersItem_BitwardenSensitiveInformation,
    UrlBlockParametersItem_Context,
    UrlBlockParametersItem_Credential,
    UrlBlockParametersItem_Output,
    UrlBlockParametersItem_Workflow,
)
from .validation_block import ValidationBlock
from .validation_block_data_schema import ValidationBlockDataSchema
from .validation_block_parameters_item import (
    ValidationBlockParametersItem,
    ValidationBlockParametersItem_AwsSecret,
    ValidationBlockParametersItem_BitwardenCreditCardData,
    ValidationBlockParametersItem_BitwardenLoginCredential,
    ValidationBlockParametersItem_BitwardenSensitiveInformation,
    ValidationBlockParametersItem_Context,
    ValidationBlockParametersItem_Credential,
    ValidationBlockParametersItem_Output,
    ValidationBlockParametersItem_Workflow,
)
from .validation_error import ValidationError
from .validation_error_loc_item import ValidationErrorLocItem
from .value import Value
from .wait_block import WaitBlock
from .wait_block_parameters_item import (
    WaitBlockParametersItem,
    WaitBlockParametersItem_AwsSecret,
    WaitBlockParametersItem_BitwardenCreditCardData,
    WaitBlockParametersItem_BitwardenLoginCredential,
    WaitBlockParametersItem_BitwardenSensitiveInformation,
    WaitBlockParametersItem_Context,
    WaitBlockParametersItem_Credential,
    WaitBlockParametersItem_Output,
    WaitBlockParametersItem_Workflow,
)
from .workflow import Workflow
from .workflow_definition import WorkflowDefinition
from .workflow_definition_blocks_item import (
    WorkflowDefinitionBlocksItem,
    WorkflowDefinitionBlocksItem_Action,
    WorkflowDefinitionBlocksItem_Code,
    WorkflowDefinitionBlocksItem_DownloadToS3,
    WorkflowDefinitionBlocksItem_Extraction,
    WorkflowDefinitionBlocksItem_FileDownload,
    WorkflowDefinitionBlocksItem_FileUpload,
    WorkflowDefinitionBlocksItem_FileUrlParser,
    WorkflowDefinitionBlocksItem_ForLoop,
    WorkflowDefinitionBlocksItem_GotoUrl,
    WorkflowDefinitionBlocksItem_Login,
    WorkflowDefinitionBlocksItem_Navigation,
    WorkflowDefinitionBlocksItem_PdfParser,
    WorkflowDefinitionBlocksItem_SendEmail,
    WorkflowDefinitionBlocksItem_Task,
    WorkflowDefinitionBlocksItem_TaskV2,
    WorkflowDefinitionBlocksItem_TextPrompt,
    WorkflowDefinitionBlocksItem_UploadToS3,
    WorkflowDefinitionBlocksItem_Validation,
    WorkflowDefinitionBlocksItem_Wait,
)
from .workflow_definition_parameters_item import (
    WorkflowDefinitionParametersItem,
    WorkflowDefinitionParametersItem_AwsSecret,
    WorkflowDefinitionParametersItem_BitwardenCreditCardData,
    WorkflowDefinitionParametersItem_BitwardenLoginCredential,
    WorkflowDefinitionParametersItem_BitwardenSensitiveInformation,
    WorkflowDefinitionParametersItem_Context,
    WorkflowDefinitionParametersItem_Credential,
    WorkflowDefinitionParametersItem_Output,
    WorkflowDefinitionParametersItem_Workflow,
)
from .workflow_parameter import WorkflowParameter
from .workflow_parameter_default_value import WorkflowParameterDefaultValue
from .workflow_parameter_type import WorkflowParameterType
from .workflow_run_request import WorkflowRunRequest
from .workflow_run_response import WorkflowRunResponse
from .workflow_status import WorkflowStatus

__all__ = [
    "ActionBlock",
    "ActionBlockDataSchema",
    "ActionBlockParametersItem",
    "ActionBlockParametersItem_AwsSecret",
    "ActionBlockParametersItem_BitwardenCreditCardData",
    "ActionBlockParametersItem_BitwardenLoginCredential",
    "ActionBlockParametersItem_BitwardenSensitiveInformation",
    "ActionBlockParametersItem_Context",
    "ActionBlockParametersItem_Credential",
    "ActionBlockParametersItem_Output",
    "ActionBlockParametersItem_Workflow",
    "AwsSecretParameter",
    "BitwardenCreditCardDataParameter",
    "BitwardenLoginCredentialParameter",
    "BitwardenSensitiveInformationParameter",
    "BrowserSessionResponse",
    "CodeBlock",
    "CodeBlockParametersItem",
    "CodeBlockParametersItem_AwsSecret",
    "CodeBlockParametersItem_BitwardenCreditCardData",
    "CodeBlockParametersItem_BitwardenLoginCredential",
    "CodeBlockParametersItem_BitwardenSensitiveInformation",
    "CodeBlockParametersItem_Context",
    "CodeBlockParametersItem_Credential",
    "CodeBlockParametersItem_Output",
    "CodeBlockParametersItem_Workflow",
    "ContextParameter",
    "CredentialParameter",
    "DownloadToS3Block",
    "ExtractionBlock",
    "ExtractionBlockDataSchema",
    "ExtractionBlockParametersItem",
    "ExtractionBlockParametersItem_AwsSecret",
    "ExtractionBlockParametersItem_BitwardenCreditCardData",
    "ExtractionBlockParametersItem_BitwardenLoginCredential",
    "ExtractionBlockParametersItem_BitwardenSensitiveInformation",
    "ExtractionBlockParametersItem_Context",
    "ExtractionBlockParametersItem_Credential",
    "ExtractionBlockParametersItem_Output",
    "ExtractionBlockParametersItem_Workflow",
    "FileDownloadBlock",
    "FileDownloadBlockDataSchema",
    "FileDownloadBlockParametersItem",
    "FileDownloadBlockParametersItem_AwsSecret",
    "FileDownloadBlockParametersItem_BitwardenCreditCardData",
    "FileDownloadBlockParametersItem_BitwardenLoginCredential",
    "FileDownloadBlockParametersItem_BitwardenSensitiveInformation",
    "FileDownloadBlockParametersItem_Context",
    "FileDownloadBlockParametersItem_Credential",
    "FileDownloadBlockParametersItem_Output",
    "FileDownloadBlockParametersItem_Workflow",
    "FileInfo",
    "FileParserBlock",
    "FileStorageType",
    "FileType",
    "FileUploadBlock",
    "ForLoopBlock",
    "ForLoopBlockLoopBlocksItem",
    "ForLoopBlockLoopBlocksItem_Action",
    "ForLoopBlockLoopBlocksItem_Code",
    "ForLoopBlockLoopBlocksItem_DownloadToS3",
    "ForLoopBlockLoopBlocksItem_Extraction",
    "ForLoopBlockLoopBlocksItem_FileDownload",
    "ForLoopBlockLoopBlocksItem_FileUpload",
    "ForLoopBlockLoopBlocksItem_FileUrlParser",
    "ForLoopBlockLoopBlocksItem_ForLoop",
    "ForLoopBlockLoopBlocksItem_GotoUrl",
    "ForLoopBlockLoopBlocksItem_Login",
    "ForLoopBlockLoopBlocksItem_Navigation",
    "ForLoopBlockLoopBlocksItem_PdfParser",
    "ForLoopBlockLoopBlocksItem_SendEmail",
    "ForLoopBlockLoopBlocksItem_Task",
    "ForLoopBlockLoopBlocksItem_TaskV2",
    "ForLoopBlockLoopBlocksItem_TextPrompt",
    "ForLoopBlockLoopBlocksItem_UploadToS3",
    "ForLoopBlockLoopBlocksItem_Validation",
    "ForLoopBlockLoopBlocksItem_Wait",
    "ForLoopBlockLoopOver",
    "ForLoopBlockLoopOver_AwsSecret",
    "ForLoopBlockLoopOver_BitwardenCreditCardData",
    "ForLoopBlockLoopOver_BitwardenLoginCredential",
    "ForLoopBlockLoopOver_BitwardenSensitiveInformation",
    "ForLoopBlockLoopOver_Context",
    "ForLoopBlockLoopOver_Credential",
    "ForLoopBlockLoopOver_Output",
    "ForLoopBlockLoopOver_Workflow",
    "HttpValidationError",
    "LoginBlock",
    "LoginBlockDataSchema",
    "LoginBlockParametersItem",
    "LoginBlockParametersItem_AwsSecret",
    "LoginBlockParametersItem_BitwardenCreditCardData",
    "LoginBlockParametersItem_BitwardenLoginCredential",
    "LoginBlockParametersItem_BitwardenSensitiveInformation",
    "LoginBlockParametersItem_Context",
    "LoginBlockParametersItem_Credential",
    "LoginBlockParametersItem_Output",
    "LoginBlockParametersItem_Workflow",
    "NavigationBlock",
    "NavigationBlockDataSchema",
    "NavigationBlockParametersItem",
    "NavigationBlockParametersItem_AwsSecret",
    "NavigationBlockParametersItem_BitwardenCreditCardData",
    "NavigationBlockParametersItem_BitwardenLoginCredential",
    "NavigationBlockParametersItem_BitwardenSensitiveInformation",
    "NavigationBlockParametersItem_Context",
    "NavigationBlockParametersItem_Credential",
    "NavigationBlockParametersItem_Output",
    "NavigationBlockParametersItem_Workflow",
    "Output",
    "OutputParameter",
    "PdfParserBlock",
    "ProxyLocation",
    "RunEngine",
    "RunStatus",
    "SendEmailBlock",
    "Source",
    "Source_AwsSecret",
    "Source_BitwardenCreditCardData",
    "Source_BitwardenLoginCredential",
    "Source_BitwardenSensitiveInformation",
    "Source_Context",
    "Source_Credential",
    "Source_Output",
    "Source_Workflow",
    "TaskBlock",
    "TaskBlockDataSchema",
    "TaskBlockParametersItem",
    "TaskBlockParametersItem_AwsSecret",
    "TaskBlockParametersItem_BitwardenCreditCardData",
    "TaskBlockParametersItem_BitwardenLoginCredential",
    "TaskBlockParametersItem_BitwardenSensitiveInformation",
    "TaskBlockParametersItem_Context",
    "TaskBlockParametersItem_Credential",
    "TaskBlockParametersItem_Output",
    "TaskBlockParametersItem_Workflow",
    "TaskRunRequest",
    "TaskRunRequestDataExtractionSchema",
    "TaskRunResponse",
    "TaskV2Block",
    "TextPromptBlock",
    "TextPromptBlockParametersItem",
    "TextPromptBlockParametersItem_AwsSecret",
    "TextPromptBlockParametersItem_BitwardenCreditCardData",
    "TextPromptBlockParametersItem_BitwardenLoginCredential",
    "TextPromptBlockParametersItem_BitwardenSensitiveInformation",
    "TextPromptBlockParametersItem_Context",
    "TextPromptBlockParametersItem_Credential",
    "TextPromptBlockParametersItem_Output",
    "TextPromptBlockParametersItem_Workflow",
    "UploadToS3Block",
    "UrlBlock",
    "UrlBlockDataSchema",
    "UrlBlockParametersItem",
    "UrlBlockParametersItem_AwsSecret",
    "UrlBlockParametersItem_BitwardenCreditCardData",
    "UrlBlockParametersItem_BitwardenLoginCredential",
    "UrlBlockParametersItem_BitwardenSensitiveInformation",
    "UrlBlockParametersItem_Context",
    "UrlBlockParametersItem_Credential",
    "UrlBlockParametersItem_Output",
    "UrlBlockParametersItem_Workflow",
    "ValidationBlock",
    "ValidationBlockDataSchema",
    "ValidationBlockParametersItem",
    "ValidationBlockParametersItem_AwsSecret",
    "ValidationBlockParametersItem_BitwardenCreditCardData",
    "ValidationBlockParametersItem_BitwardenLoginCredential",
    "ValidationBlockParametersItem_BitwardenSensitiveInformation",
    "ValidationBlockParametersItem_Context",
    "ValidationBlockParametersItem_Credential",
    "ValidationBlockParametersItem_Output",
    "ValidationBlockParametersItem_Workflow",
    "ValidationError",
    "ValidationErrorLocItem",
    "Value",
    "WaitBlock",
    "WaitBlockParametersItem",
    "WaitBlockParametersItem_AwsSecret",
    "WaitBlockParametersItem_BitwardenCreditCardData",
    "WaitBlockParametersItem_BitwardenLoginCredential",
    "WaitBlockParametersItem_BitwardenSensitiveInformation",
    "WaitBlockParametersItem_Context",
    "WaitBlockParametersItem_Credential",
    "WaitBlockParametersItem_Output",
    "WaitBlockParametersItem_Workflow",
    "Workflow",
    "WorkflowDefinition",
    "WorkflowDefinitionBlocksItem",
    "WorkflowDefinitionBlocksItem_Action",
    "WorkflowDefinitionBlocksItem_Code",
    "WorkflowDefinitionBlocksItem_DownloadToS3",
    "WorkflowDefinitionBlocksItem_Extraction",
    "WorkflowDefinitionBlocksItem_FileDownload",
    "WorkflowDefinitionBlocksItem_FileUpload",
    "WorkflowDefinitionBlocksItem_FileUrlParser",
    "WorkflowDefinitionBlocksItem_ForLoop",
    "WorkflowDefinitionBlocksItem_GotoUrl",
    "WorkflowDefinitionBlocksItem_Login",
    "WorkflowDefinitionBlocksItem_Navigation",
    "WorkflowDefinitionBlocksItem_PdfParser",
    "WorkflowDefinitionBlocksItem_SendEmail",
    "WorkflowDefinitionBlocksItem_Task",
    "WorkflowDefinitionBlocksItem_TaskV2",
    "WorkflowDefinitionBlocksItem_TextPrompt",
    "WorkflowDefinitionBlocksItem_UploadToS3",
    "WorkflowDefinitionBlocksItem_Validation",
    "WorkflowDefinitionBlocksItem_Wait",
    "WorkflowDefinitionParametersItem",
    "WorkflowDefinitionParametersItem_AwsSecret",
    "WorkflowDefinitionParametersItem_BitwardenCreditCardData",
    "WorkflowDefinitionParametersItem_BitwardenLoginCredential",
    "WorkflowDefinitionParametersItem_BitwardenSensitiveInformation",
    "WorkflowDefinitionParametersItem_Context",
    "WorkflowDefinitionParametersItem_Credential",
    "WorkflowDefinitionParametersItem_Output",
    "WorkflowDefinitionParametersItem_Workflow",
    "WorkflowParameter",
    "WorkflowParameterDefaultValue",
    "WorkflowParameterType",
    "WorkflowRunRequest",
    "WorkflowRunResponse",
    "WorkflowStatus",
]



"""Add suggested_title to task_generations table

Revision ID: 6de11b2be7c8
Revises: 8f237f00faeb
Create Date: 2024-08-23 20:12:13.426060+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "6de11b2be7c8"
down_revision: Union[str, None] = "8f237f00faeb"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("task_generations", sa.Column("suggested_title", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("task_generations", "suggested_title")
    # ### end Alembic commands ###



"""update task_generation table - use user_prompt_hash as the index of a user prompt

Revision ID: 0de9150bc624
Revises: 6de11b2be7c8
Create Date: 2024-09-03 03:56:58.352307+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "0de9150bc624"
down_revision: Union[str, None] = "6de11b2be7c8"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("task_generations", sa.Column("user_prompt_hash", sa.String(), nullable=True))
    op.add_column("task_generations", sa.Column("source_task_generation_id", sa.String(), nullable=True))
    op.drop_index("ix_task_generations_user_prompt", table_name="task_generations")
    op.create_index(
        op.f("ix_task_generations_source_task_generation_id"),
        "task_generations",
        ["source_task_generation_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_task_generations_user_prompt_hash"), "task_generations", ["user_prompt_hash"], unique=False
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_task_generations_user_prompt_hash"), table_name="task_generations")
    op.drop_index(op.f("ix_task_generations_source_task_generation_id"), table_name="task_generations")
    op.create_index("ix_task_generations_user_prompt", "task_generations", ["user_prompt"], unique=False)
    op.drop_column("task_generations", "source_task_generation_id")
    op.drop_column("task_generations", "user_prompt_hash")
    # ### end Alembic commands ###



import structlog

from skyvern.exceptions import CachedActionPlanError
from skyvern.forge import app
from skyvern.forge.prompts import prompt_engine
from skyvern.forge.sdk.models import Step
from skyvern.forge.sdk.schemas.tasks import Task
from skyvern.webeye.actions.actions import Action, ActionStatus, ActionType
from skyvern.webeye.scraper.scraper import ScrapedPage

LOG = structlog.get_logger()


async def retrieve_action_plan(task: Task, step: Step, scraped_page: ScrapedPage) -> list[Action]:
    try:
        return await _retrieve_action_plan(task, step, scraped_page)
    except Exception as e:
        LOG.exception("Failed to retrieve action plan", exception=e)
        return []


async def _retrieve_action_plan(task: Task, step: Step, scraped_page: ScrapedPage) -> list[Action]:
    # V0: use the previous action plan if there is a completed task with the same url and navigation goal
    # get completed task with the same url and navigation goal
    # TODO(kerem): don't use step_order, get all the previous actions instead
    cached_actions = await app.DATABASE.retrieve_action_plan(task=task)
    if not cached_actions:
        LOG.info("No cached actions found for the task, fallback to no-cache mode")
        return []

    # Get the existing actions for this task from the database. Then find the actions that are already executed by looking at
    # the source_action_id field for this task's actions.
    previous_actions = await app.DATABASE.get_previous_actions_for_task(task_id=task.task_id)

    executed_cached_actions = []
    remaining_cached_actions = []
    action_matching_complete = False
    if previous_actions:
        for idx, cached_action in enumerate(cached_actions):
            if not action_matching_complete:
                should_be_matching_action = previous_actions[idx]
                if not should_be_matching_action.source_action_id:
                    # If there is an action without a source_action_id, it means we already went back to no-cache mode
                    # and we should not try to reuse the previous action plan since it's not possible to determine which
                    # action we should execute next
                    return []

                action_id_to_match = (
                    cached_action.source_action_id if cached_action.source_action_id else cached_action.action_id
                )
                if should_be_matching_action.source_action_id == action_id_to_match:
                    executed_cached_actions.append(cached_action)
                    if idx == len(previous_actions) - 1:
                        # If we've reached the end of the previous actions, we've completed matching.
                        action_matching_complete = True
                else:
                    # If we've reached an action that doesn't match the source_action_id of the previous actions,
                    # we've completed matching.
                    action_matching_complete = True
                    remaining_cached_actions.append(cached_action)
            else:
                remaining_cached_actions.append(cached_action)
    else:
        remaining_cached_actions = cached_actions
        action_matching_complete = True

    # For any remaining cached action,
    # check if the element hash exists in the current scraped page. Add them to a list until we can't find a match. Always keep the
    # actions without an element hash.

    cached_actions_to_execute: list[Action] = []
    found_element_with_no_hash = False
    for cached_action in remaining_cached_actions:
        # The actions without an element hash: TerminateAction CompleteAction NullAction SolveCaptchaAction WaitAction
        # For these actions, we can't check if the element hash exists in the current scraped page.
        # For that reason, we're going to make sure they're executed always as the first action in each step.
        if not cached_action.skyvern_element_hash:
            if not found_element_with_no_hash and len(cached_actions_to_execute) > 0:
                # If we've already added actions with element hashes to the list before we encounter an action without an element hash,
                # we need to execute the actions we already added first. We want the actions without an element hash
                # to be executed as the first actions in each step. We're ok with executing multiple actions without an element hash
                # in a row, but we want them to be executed in a new step after we wait & scrape the page again.
                break
            cached_actions_to_execute.append(cached_action)
            found_element_with_no_hash = True
            continue

        matching_element_ids = scraped_page.hash_to_element_ids.get(cached_action.skyvern_element_hash)
        if matching_element_ids and len(matching_element_ids) == 1:
            cached_actions_to_execute.append(cached_action)
            continue
        # After this point, we can't continue adding actions to the plan, so we break and continue with what we have.
        # Because this action has either no hash-match or multiple hash-matches, we can't continue.
        elif matching_element_ids and len(matching_element_ids) > 1:
            LOG.warning(
                "Found multiple elements with the same hash, stop matching",
                element_hash=cached_action.skyvern_element_hash,
                element_ids=matching_element_ids,
            )
            break
        else:
            LOG.warning("No element found with the hash", element_hash=cached_action.skyvern_element_hash)
            break

    # If there are no items in the list we just built, we need to revert back to no-cache mode. Return empty list.
    if not cached_actions_to_execute:
        return []

    LOG.info("Found cached actions to execute", actions=cached_actions_to_execute)

    actions: list[Action] = []
    for idx, cached_action in enumerate(cached_actions_to_execute):
        updated_action = cached_action.model_copy()
        updated_action.status = ActionStatus.pending
        updated_action.source_action_id = (
            cached_action.source_action_id if cached_action.source_action_id else cached_action.action_id
        )
        updated_action.workflow_run_id = task.workflow_run_id
        updated_action.task_id = task.task_id
        updated_action.step_id = step.step_id
        updated_action.step_order = step.order
        updated_action.action_order = idx
        # Reset the action response to None so we don't use the previous answers
        updated_action.response = None

        # Update the element id with the element id from the current scraped page, matched by element hash
        if cached_action.skyvern_element_hash:
            matching_element_ids = scraped_page.hash_to_element_ids.get(cached_action.skyvern_element_hash)
            if matching_element_ids and len(matching_element_ids) == 1:
                matching_element_id = matching_element_ids[0]
                updated_action.element_id = matching_element_id
                updated_action.skyvern_element_data = scraped_page.id_to_element_dict.get(matching_element_id)
            else:
                raise CachedActionPlanError(
                    "All elements with either no hash or multiple hashes should have been already filtered out"
                )

        actions.append(updated_action)

    # Check for unsupported actions before personalizing the actions
    # Classify the supported actions into two groups:
    # 1. Actions that can be cached with a query
    # 2. Actions that can be cached without a query
    # We'll use this classification to determine if we should continue with caching or fallback to no-cache mode
    check_for_unsupported_actions(actions)

    personalized_actions = await personalize_actions(task=task, step=step, scraped_page=scraped_page, actions=actions)

    LOG.info("Personalized cached actions are ready", actions=personalized_actions)
    return personalized_actions


async def personalize_actions(
    task: Task,
    step: Step,
    actions: list[Action],
    scraped_page: ScrapedPage,
) -> list[Action]:
    queries_and_answers: dict[str, str | None] = {action.intention: None for action in actions if action.intention}

    answered_queries: dict[str, str] = {}
    if queries_and_answers:
        # Call LLM to get answers for the queries only if there are queries to answer
        answered_queries = await get_user_detail_answers(
            task=task, step=step, scraped_page=scraped_page, queries_and_answers=queries_and_answers
        )

    personalized_actions = []
    for action in actions:
        query = action.intention
        if query and (personalized_answer := answered_queries.get(query)):
            current_personized_actions = await personalize_action(
                action, query, personalized_answer, task, step, scraped_page
            )
            personalized_actions.extend(current_personized_actions)
        else:
            personalized_actions.append(action)

    return personalized_actions


async def get_user_detail_answers(
    task: Task, step: Step, scraped_page: ScrapedPage, queries_and_answers: dict[str, str | None]
) -> dict[str, str]:
    try:
        question_answering_prompt = prompt_engine.load_prompt(
            "answer-user-detail-questions",
            navigation_goal=task.navigation_goal,
            navigation_payload=task.navigation_payload,
            queries_and_answers=queries_and_answers,
        )

        llm_response = await app.SECONDARY_LLM_API_HANDLER(
            prompt=question_answering_prompt, step=step, screenshots=None, prompt_name="answer-user-detail-questions"
        )
        return llm_response
    except Exception as e:
        LOG.exception("Failed to get user detail answers", exception=e)
        # TODO: custom exception so we can fallback to no-cache mode by catching it
        raise e


async def personalize_action(
    action: Action,
    query: str,
    answer: str,
    task: Task,
    step: Step,
    scraped_page: ScrapedPage,
) -> list[Action]:
    action.intention = query
    action.response = answer

    if action.action_type == ActionType.INPUT_TEXT:
        action.text = answer
        if not answer:
            return []
    elif action.action_type == ActionType.UPLOAD_FILE:
        action.file_url = answer
    elif action.action_type == ActionType.CLICK:
        # TODO: we only use cached action.intention. send the intention, navigation payload + navigation goal, html
        # to small llm and make a decision of which elements to click. Not clicking anything is also an option here
        return [action]
    elif action.action_type == ActionType.SELECT_OPTION:
        # TODO: send the selection action with the original/previous option value. Our current selection agent
        # is already able to handle it
        return [action]
    elif action.action_type in [
        ActionType.COMPLETE,
        ActionType.WAIT,
        ActionType.SOLVE_CAPTCHA,
        ActionType.NULL_ACTION,
    ]:
        return [action]
    elif action.action_type == ActionType.TERMINATE:
        return []
    else:
        raise CachedActionPlanError(
            f"Unsupported action type for personalization, fallback to no-cache mode: {action.action_type}"
        )

    return [action]


def check_for_unsupported_actions(actions: list[Action]) -> None:
    supported_actions = [ActionType.INPUT_TEXT, ActionType.WAIT, ActionType.CLICK, ActionType.COMPLETE]
    supported_actions_with_query = [ActionType.INPUT_TEXT]
    for action in actions:
        query = action.intention
        if action.action_type not in supported_actions:
            raise CachedActionPlanError(
                f"This action type does not support caching: {action.action_type}, fallback to no-cache mode"
            )
        if query and action.action_type not in supported_actions_with_query:
            raise CachedActionPlanError(
                f"This action type does not support caching with a query: {action.action_type}, fallback to no-cache mode"
            )



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations
from ...core.pydantic_utilities import UniversalBaseModel
import typing
from ...types.run_status import RunStatus
from ...types.output import Output
from ...types.file_info import FileInfo
import datetime as dt
from ...types.task_run_request import TaskRunRequest
from ...core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic
from ...types.workflow_run_request import WorkflowRunRequest


class AgentGetRunResponse_TaskV1(UniversalBaseModel):
    run_type: typing.Literal["task_v1"] = "task_v1"
    run_id: str
    status: RunStatus
    output: typing.Optional[Output] = None
    downloaded_files: typing.Optional[typing.List[FileInfo]] = None
    recording_url: typing.Optional[str] = None
    failure_reason: typing.Optional[str] = None
    created_at: dt.datetime
    modified_at: dt.datetime
    run_request: typing.Optional[TaskRunRequest] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class AgentGetRunResponse_TaskV2(UniversalBaseModel):
    run_type: typing.Literal["task_v2"] = "task_v2"
    run_id: str
    status: RunStatus
    output: typing.Optional[Output] = None
    downloaded_files: typing.Optional[typing.List[FileInfo]] = None
    recording_url: typing.Optional[str] = None
    failure_reason: typing.Optional[str] = None
    created_at: dt.datetime
    modified_at: dt.datetime
    run_request: typing.Optional[TaskRunRequest] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class AgentGetRunResponse_WorkflowRun(UniversalBaseModel):
    run_type: typing.Literal["workflow_run"] = "workflow_run"
    run_id: str
    status: RunStatus
    output: typing.Optional[Output] = None
    downloaded_files: typing.Optional[typing.List[FileInfo]] = None
    recording_url: typing.Optional[str] = None
    failure_reason: typing.Optional[str] = None
    created_at: dt.datetime
    modified_at: dt.datetime
    run_request: typing.Optional[WorkflowRunRequest] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


AgentGetRunResponse = typing.Union[
    AgentGetRunResponse_TaskV1, AgentGetRunResponse_TaskV2, AgentGetRunResponse_WorkflowRun
]



import platform
import subprocess
from pathlib import Path
from typing import Optional

from alembic import command
from alembic.config import Config
from skyvern.constants import REPO_ROOT_DIR


def migrate_db() -> None:
    alembic_cfg = Config()
    path = f"{REPO_ROOT_DIR}/alembic"
    alembic_cfg.set_main_option("script_location", path)
    command.upgrade(alembic_cfg, "head")


def detect_os() -> str:
    """
    Detects the operating system.

    Returns:
        str: The name of the OS in lowercase.
             Returns 'wsl' for Windows Subsystem for Linux,
             'linux' for native Linux,
             or the lowercase name of other platforms (e.g., 'windows', 'darwin').
    """
    system = platform.system()
    if system == "Linux":
        try:
            with open("/proc/version", "r") as f:
                version_info = f.read().lower()
                if "microsoft" in version_info:
                    return "wsl"
        except Exception:
            pass
        return "linux"
    else:
        return system.lower()


def get_windows_appdata_roaming() -> Optional[Path]:
    """
    Retrieves the Windows 'AppData\\Roaming' directory path from WSL.

    Returns:
        Optional[Path]: A Path object representing the translated Linux-style path
                        to the Windows AppData\\Roaming folder, or None if retrieval fails.
    """
    try:
        output = (
            subprocess.check_output(
                ["powershell.exe", "-NoProfile", "-Command", "[Environment]::GetFolderPath('ApplicationData')"],
                stderr=subprocess.DEVNULL,
            )
            .decode("utf-8")
            .strip()
        )
        linux_path = "/mnt/" + output[0].lower() + output[2:].replace("\\", "/")
        return Path(linux_path)
    except Exception:
        return None



from fastapi import Depends, HTTPException
from fastapi.responses import ORJSONResponse

from skyvern import analytics
from skyvern.forge import app
from skyvern.forge.sdk.routes.routers import base_router, legacy_base_router
from skyvern.forge.sdk.schemas.organizations import Organization
from skyvern.forge.sdk.services import org_auth_service
from skyvern.webeye.schemas import BrowserSessionResponse


@base_router.get(
    "/browser_sessions/{browser_session_id}",
    response_model=BrowserSessionResponse,
    tags=["Browser Sessions"],
    openapi_extra={
        "x-fern-sdk-group-name": "browser_session",
        "x-fern-sdk-method-name": "get_browser_session",
    },
    description="Get details about a specific browser session by ID",
    summary="Get browser session details",
    responses={
        200: {"description": "Successfully retrieved browser session details"},
        404: {"description": "Browser session not found"},
        401: {"description": "Unauthorized - Invalid or missing authentication"},
    },
)
@legacy_base_router.get(
    "/browser_sessions/{browser_session_id}",
    response_model=BrowserSessionResponse,
    tags=["session"],
    openapi_extra={
        "x-fern-sdk-group-name": "session",
        "x-fern-sdk-method-name": "get_browser_session",
    },
)
@legacy_base_router.get(
    "/browser_sessions/{browser_session_id}/",
    response_model=BrowserSessionResponse,
    include_in_schema=False,
)
async def get_browser_session(
    browser_session_id: str,
    current_org: Organization = Depends(org_auth_service.get_current_org),
) -> BrowserSessionResponse:
    analytics.capture("skyvern-oss-agent-browser-session-get")
    browser_session = await app.PERSISTENT_SESSIONS_MANAGER.get_session(
        browser_session_id,
        current_org.organization_id,
    )
    if not browser_session:
        raise HTTPException(status_code=404, detail=f"Browser session {browser_session_id} not found")
    return BrowserSessionResponse.from_browser_session(browser_session)


@base_router.get(
    "/browser_sessions",
    response_model=list[BrowserSessionResponse],
    tags=["Browser Sessions"],
    openapi_extra={
        "x-fern-sdk-group-name": "browser_session",
        "x-fern-sdk-method-name": "get_browser_sessions",
    },
    description="Get all active browser sessions for the organization",
    summary="Get all active browser sessions",
    responses={
        200: {"description": "Successfully retrieved all active browser sessions"},
        401: {"description": "Unauthorized - Invalid or missing authentication"},
    },
)
@legacy_base_router.get(
    "/browser_sessions",
    response_model=list[BrowserSessionResponse],
    tags=["session"],
    openapi_extra={
        "x-fern-sdk-group-name": "browser_sessions",
        "x-fern-sdk-method-name": "get_browser_sessions",
    },
)
@legacy_base_router.get(
    "/browser_sessions/",
    response_model=list[BrowserSessionResponse],
    include_in_schema=False,
)
async def get_browser_sessions(
    current_org: Organization = Depends(org_auth_service.get_current_org),
) -> list[BrowserSessionResponse]:
    """Get all active browser sessions for the organization"""
    analytics.capture("skyvern-oss-agent-browser-sessions-get")
    browser_sessions = await app.PERSISTENT_SESSIONS_MANAGER.get_active_sessions(current_org.organization_id)
    return [BrowserSessionResponse.from_browser_session(browser_session) for browser_session in browser_sessions]


@base_router.post(
    "/browser_sessions",
    response_model=BrowserSessionResponse,
    tags=["Browser Sessions"],
    openapi_extra={
        "x-fern-sdk-group-name": "browser_session",
        "x-fern-sdk-method-name": "create_browser_session",
    },
    description="Create a new browser session",
    summary="Create a new browser session",
    responses={
        200: {"description": "Successfully created browser session"},
        401: {"description": "Unauthorized - Invalid or missing authentication"},
    },
)
@legacy_base_router.post(
    "/browser_sessions",
    response_model=BrowserSessionResponse,
    tags=["Browser Sessions"],
    openapi_extra={
        "x-fern-sdk-group-name": "session",
        "x-fern-sdk-method-name": "create_browser_session",
    },
)
@legacy_base_router.post(
    "/browser_sessions/",
    response_model=BrowserSessionResponse,
    include_in_schema=False,
)
async def create_browser_session(
    current_org: Organization = Depends(org_auth_service.get_current_org),
) -> BrowserSessionResponse:
    browser_session = await app.PERSISTENT_SESSIONS_MANAGER.create_session(current_org.organization_id)
    return BrowserSessionResponse.from_browser_session(browser_session)


@base_router.post(
    "/browser_sessions/{browser_session_id}/close",
    tags=["Browser Sessions"],
    openapi_extra={
        "x-fern-sdk-group-name": "browser_session",
        "x-fern-sdk-method-name": "close_browser_session",
    },
    description="Close a browser session",
    summary="Close a browser session",
    responses={
        200: {"description": "Successfully closed browser session"},
        401: {"description": "Unauthorized - Invalid or missing authentication"},
    },
)
@legacy_base_router.post(
    "/browser_sessions/{browser_session_id}/close",
    tags=["Browser Sessions"],
    openapi_extra={
        "x-fern-sdk-group-name": "browser_session",
        "x-fern-sdk-method-name": "close_browser_session",
    },
)
@legacy_base_router.post(
    "/browser_sessions/{browser_session_id}/close/",
    include_in_schema=False,
)
async def close_browser_session(
    browser_session_id: str,
    current_org: Organization = Depends(org_auth_service.get_current_org),
) -> ORJSONResponse:
    await app.PERSISTENT_SESSIONS_MANAGER.close_session(current_org.organization_id, browser_session_id)
    return ORJSONResponse(
        content={"message": "Browser session closed"},
        status_code=200,
        media_type="application/json",
    )



"""introduce persistent browser sessions

Revision ID: 282b0548d443
Revises: 411dd89f3df9
Create Date: 2024-12-17 18:41:30.400052+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "282b0548d443"
down_revision: Union[str, None] = "411dd89f3df9"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "persistent_browser_sessions",
        sa.Column("persistent_browser_session_id", sa.String(), nullable=False),
        sa.Column("organization_id", sa.String(), nullable=False),
        sa.Column("runnable_type", sa.String(), nullable=False),
        sa.Column("runnable_id", sa.String(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("modified_at", sa.DateTime(), nullable=False),
        sa.Column("deleted_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["organization_id"],
            ["organizations.organization_id"],
        ),
        sa.PrimaryKeyConstraint("persistent_browser_session_id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("persistent_browser_sessions")
    # ### end Alembic commands ###



from skyvern.forge.sdk.executor.async_executor import AsyncExecutor, BackgroundTaskExecutor


class AsyncExecutorFactory:
    __instance: AsyncExecutor = BackgroundTaskExecutor()

    @staticmethod
    def set_executor(executor: AsyncExecutor) -> None:
        AsyncExecutorFactory.__instance = executor

    @staticmethod
    def get_executor() -> AsyncExecutor:
        return AsyncExecutorFactory.__instance



import os
import shutil
from datetime import datetime
from pathlib import Path

import structlog

from skyvern.config import settings
from skyvern.forge.sdk.api.files import (
    calculate_sha256_for_file,
    get_download_dir,
    get_skyvern_temp_dir,
    parse_uri_to_path,
)
from skyvern.forge.sdk.artifact.models import Artifact, ArtifactType, LogEntityType
from skyvern.forge.sdk.artifact.storage.base import FILE_EXTENTSION_MAP, BaseStorage
from skyvern.forge.sdk.models import Step
from skyvern.forge.sdk.schemas.ai_suggestions import AISuggestion
from skyvern.forge.sdk.schemas.files import FileInfo
from skyvern.forge.sdk.schemas.task_v2 import TaskV2, Thought
from skyvern.forge.sdk.schemas.workflow_runs import WorkflowRunBlock

LOG = structlog.get_logger()


class LocalStorage(BaseStorage):
    def __init__(self, artifact_path: str = settings.ARTIFACT_STORAGE_PATH) -> None:
        self.artifact_path = artifact_path

    def build_uri(self, artifact_id: str, step: Step, artifact_type: ArtifactType) -> str:
        file_ext = FILE_EXTENTSION_MAP[artifact_type]
        return f"file://{self.artifact_path}/{step.task_id}/{step.order:02d}_{step.retry_index}_{step.step_id}/{datetime.utcnow().isoformat()}_{artifact_id}_{artifact_type}.{file_ext}"

    async def retrieve_global_workflows(self) -> list[str]:
        file_path = Path(f"{self.artifact_path}/{settings.ENV}/global_workflows.txt")
        self._create_directories_if_not_exists(file_path)
        if not file_path.exists():
            return []
        try:
            with open(file_path, "r") as f:
                return [line.strip() for line in f.readlines() if line.strip()]
        except Exception:
            return []

    def build_log_uri(self, log_entity_type: LogEntityType, log_entity_id: str, artifact_type: ArtifactType) -> str:
        file_ext = FILE_EXTENTSION_MAP[artifact_type]
        return f"file://{self.artifact_path}/logs/{log_entity_type}/{log_entity_id}/{datetime.utcnow().isoformat()}_{artifact_type}.{file_ext}"

    def build_thought_uri(self, artifact_id: str, thought: Thought, artifact_type: ArtifactType) -> str:
        file_ext = FILE_EXTENTSION_MAP[artifact_type]
        return f"file://{self.artifact_path}/{settings.ENV}/tasks/{thought.observer_cruise_id}/{thought.observer_thought_id}/{datetime.utcnow().isoformat()}_{artifact_id}_{artifact_type}.{file_ext}"

    def build_task_v2_uri(self, artifact_id: str, task_v2: TaskV2, artifact_type: ArtifactType) -> str:
        file_ext = FILE_EXTENTSION_MAP[artifact_type]
        return f"file://{self.artifact_path}/{settings.ENV}/observers/{task_v2.observer_cruise_id}/{datetime.utcnow().isoformat()}_{artifact_id}_{artifact_type}.{file_ext}"

    def build_workflow_run_block_uri(
        self, artifact_id: str, workflow_run_block: WorkflowRunBlock, artifact_type: ArtifactType
    ) -> str:
        file_ext = FILE_EXTENTSION_MAP[artifact_type]
        return f"file://{self.artifact_path}/{settings.ENV}/workflow_runs/{workflow_run_block.workflow_run_id}/{workflow_run_block.workflow_run_block_id}/{datetime.utcnow().isoformat()}_{artifact_id}_{artifact_type}.{file_ext}"

    def build_ai_suggestion_uri(
        self, artifact_id: str, ai_suggestion: AISuggestion, artifact_type: ArtifactType
    ) -> str:
        file_ext = FILE_EXTENTSION_MAP[artifact_type]
        return f"file://{self.artifact_path}/{settings.ENV}/ai_suggestions/{ai_suggestion.ai_suggestion_id}/{datetime.utcnow().isoformat()}_{artifact_id}_{artifact_type}.{file_ext}"

    async def store_artifact(self, artifact: Artifact, data: bytes) -> None:
        file_path = None
        try:
            file_path = Path(parse_uri_to_path(artifact.uri))
            self._create_directories_if_not_exists(file_path)
            with open(file_path, "wb") as f:
                f.write(data)
        except Exception:
            LOG.exception(
                "Failed to store artifact locally.",
                file_path=file_path,
                artifact=artifact,
            )

    async def store_artifact_from_path(self, artifact: Artifact, path: str) -> None:
        file_path = None
        try:
            file_path = Path(parse_uri_to_path(artifact.uri))
            self._create_directories_if_not_exists(file_path)
            Path(path).replace(file_path)
        except Exception:
            LOG.exception(
                "Failed to store artifact locally.",
                file_path=file_path,
                artifact=artifact,
            )

    async def retrieve_artifact(self, artifact: Artifact) -> bytes | None:
        file_path = None
        try:
            file_path = parse_uri_to_path(artifact.uri)
            with open(file_path, "rb") as f:
                return f.read()
        except Exception:
            LOG.exception(
                "Failed to retrieve local artifact.",
                file_path=file_path,
                artifact=artifact,
            )
            return None

    async def get_share_link(self, artifact: Artifact) -> str:
        return artifact.uri

    async def get_share_links(self, artifacts: list[Artifact]) -> list[str]:
        return [artifact.uri for artifact in artifacts]

    async def save_streaming_file(self, organization_id: str, file_name: str) -> None:
        return

    async def get_streaming_file(self, organization_id: str, file_name: str, use_default: bool = True) -> bytes | None:
        file_path = Path(f"{get_skyvern_temp_dir()}/skyvern_screenshot.png")
        if not use_default:
            file_path = Path(f"{get_skyvern_temp_dir()}/{organization_id}/{file_name}")
        try:
            with open(file_path, "rb") as f:
                return f.read()
        except Exception:
            return None

    async def store_browser_session(self, organization_id: str, workflow_permanent_id: str, directory: str) -> None:
        stored_folder_path = Path(settings.BROWSER_SESSION_BASE_PATH) / organization_id / workflow_permanent_id
        if directory == str(stored_folder_path):
            return
        self._create_directories_if_not_exists(stored_folder_path)
        LOG.info(
            "Storing browser session locally",
            organization_id=organization_id,
            workflow_permanent_id=workflow_permanent_id,
            directory=directory,
            browser_session_path=stored_folder_path,
        )

        # Copy all files from the directory to the stored folder
        for root, _, files in os.walk(directory):
            for file in files:
                source_file_path = Path(root) / file
                relative_path = source_file_path.relative_to(directory)
                target_file_path = stored_folder_path / relative_path
                self._create_directories_if_not_exists(target_file_path)
                shutil.copy2(source_file_path, target_file_path)

    async def retrieve_browser_session(self, organization_id: str, workflow_permanent_id: str) -> str | None:
        stored_folder_path = Path(settings.BROWSER_SESSION_BASE_PATH) / organization_id / workflow_permanent_id
        if not stored_folder_path.exists():
            return None
        return str(stored_folder_path)

    async def save_downloaded_files(
        self, organization_id: str, task_id: str | None, workflow_run_id: str | None
    ) -> None:
        pass

    async def get_downloaded_files(
        self, organization_id: str, task_id: str | None, workflow_run_id: str | None
    ) -> list[FileInfo]:
        download_dir = get_download_dir(workflow_run_id=workflow_run_id, task_id=task_id)
        file_infos: list[FileInfo] = []
        files_and_folders = os.listdir(download_dir)
        for file_or_folder in files_and_folders:
            path = os.path.join(download_dir, file_or_folder)
            if os.path.isfile(path):
                # Calculate checksum for the file
                checksum = calculate_sha256_for_file(path)
                file_info = FileInfo(url=f"file://{path}", checksum=checksum, filename=file_or_folder)
                file_infos.append(file_info)
        return file_infos

    @staticmethod
    def _create_directories_if_not_exists(path_including_file_name: Path) -> None:
        path = path_including_file_name.parent
        path.mkdir(parents=True, exist_ok=True)



"""Add title, error_code_mapping, and errors to tasks

Revision ID: 82a0c686152d
Revises: 99423c1dec60
Create Date: 2024-03-13 05:18:52.674264+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "82a0c686152d"
down_revision: Union[str, None] = "99423c1dec60"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("tasks", sa.Column("title", sa.String(), nullable=True))
    op.add_column("tasks", sa.Column("error_code_mapping", sa.JSON(), nullable=True))
    # In order to add a column with a default value, we need to add the column
    # as nullable, then set the default value, then set the column to not
    op.add_column("tasks", sa.Column("errors", sa.JSON(), nullable=True))
    op.execute("UPDATE tasks SET errors = '[]'::jsonb")
    op.alter_column("tasks", "errors", nullable=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("tasks", "errors")
    op.drop_column("tasks", "error_code_mapping")
    op.drop_column("tasks", "title")
    # ### end Alembic commands ###



import abc

from skyvern.forge.sdk.schemas.organizations import Organization


class PermissionChecker(abc.ABC):
    @abc.abstractmethod
    async def check(self, organization: Organization, browser_session_id: str | None = None) -> None:
        pass


class NoopPermissionChecker(PermissionChecker):
    async def check(self, organization: Organization, browser_session_id: str | None = None) -> None:
        return



from datetime import datetime

from pydantic import BaseModel, ConfigDict


class PersistentBrowserSession(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    persistent_browser_session_id: str
    organization_id: str
    runnable_type: str | None = None
    runnable_id: str | None = None
    browser_address: str | None = None
    status: str | None = None
    created_at: datetime
    modified_at: datetime
    deleted_at: datetime | None = None



import json
from datetime import datetime
from typing import Any

import structlog
from structlog.dev import ConsoleRenderer

from skyvern.forge.skyvern_json_encoder import SkyvernJSONLogEncoder

LOG = structlog.get_logger()


class SkyvernLogEncoder:
    """Encodes Skyvern logs from JSON format to human-readable string format"""

    def __init__(self) -> None:
        self.renderer = ConsoleRenderer(
            pad_event=30,
            colors=False,
        )

    @classmethod
    def _format_value(cls, value: Any) -> str:
        return SkyvernJSONLogEncoder.dumps(value, sort_keys=True)

    @staticmethod
    def _parse_json_entry(entry: dict[str, Any]) -> dict[str, Any]:
        """Convert a JSON log entry into our standard format."""
        event = entry.get("message", entry.get("event", ""))

        clean_entry = {
            "timestamp": entry.get("timestamp", datetime.utcnow().isoformat() + "Z"),
            "level": entry.get("level", "info").lower(),
            "event": event,
        }

        for key, value in entry.items():
            if key not in ("timestamp", "level", "event", "message"):
                clean_entry[key] = SkyvernLogEncoder._format_value(value)

        return clean_entry

    @classmethod
    def encode(cls, log_entries: list[dict[str, Any]]) -> str:
        """
        Encode log entries into formatted string output using structlog's ConsoleRenderer.

        Args:
            log_entries: List of log entry dictionaries

        Returns:
            Formatted string with one log entry per line
        """
        encoder = cls()
        formatted_lines = []

        for entry in log_entries:
            try:
                if isinstance(entry, str):
                    try:
                        entry = json.loads(entry)
                    except json.JSONDecodeError:
                        entry = {"event": entry, "level": "info"}

                parsed_entry = cls._parse_json_entry(entry)

                formatted_line = encoder.renderer(None, None, parsed_entry)
                formatted_lines.append(formatted_line)

            except Exception as e:
                LOG.error("Failed to format log entry", entry=entry, error=str(e), exc_info=True)
                # Add error line to output
                error_timestamp = datetime.utcnow().isoformat() + "Z"
                error_entry = {
                    "timestamp": error_timestamp,
                    "level": "error",
                    "event": "Failed to format log entry",
                    "entry": str(entry),
                    "error": str(e),
                }
                formatted_lines.append(encoder.renderer(None, None, error_entry))

        return "\n".join(formatted_lines)



"""create totp_codes table and add task.totp_identifier

Revision ID: c5848cc524b1
Revises: c50f0aa0ef24
Create Date: 2024-09-08 21:59:56.666276+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "c5848cc524b1"
down_revision: Union[str, None] = "c50f0aa0ef24"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "totp_codes",
        sa.Column("totp_code_id", sa.String(), nullable=False),
        sa.Column("totp_identifier", sa.String(), nullable=False),
        sa.Column("organization_id", sa.String(), nullable=True),
        sa.Column("task_id", sa.String(), nullable=True),
        sa.Column("workflow_id", sa.String(), nullable=True),
        sa.Column("content", sa.String(), nullable=False),
        sa.Column("code", sa.String(), nullable=False),
        sa.Column("source", sa.String(), nullable=True),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("modified_at", sa.DateTime(), nullable=False),
        sa.Column("expired_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["organization_id"],
            ["organizations.organization_id"],
        ),
        sa.ForeignKeyConstraint(
            ["task_id"],
            ["tasks.task_id"],
        ),
        sa.ForeignKeyConstraint(
            ["workflow_id"],
            ["workflows.workflow_id"],
        ),
        sa.PrimaryKeyConstraint("totp_code_id"),
    )
    op.create_index(op.f("ix_totp_codes_created_at"), "totp_codes", ["created_at"], unique=False)
    op.create_index(op.f("ix_totp_codes_expired_at"), "totp_codes", ["expired_at"], unique=False)
    op.create_index(op.f("ix_totp_codes_totp_identifier"), "totp_codes", ["totp_identifier"], unique=False)
    op.add_column("tasks", sa.Column("totp_identifier", sa.String(), nullable=True))
    op.add_column("workflow_runs", sa.Column("totp_identifier", sa.String(), nullable=True))
    op.add_column("workflows", sa.Column("totp_identifier", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workflows", "totp_identifier")
    op.drop_column("workflow_runs", "totp_identifier")
    op.drop_column("tasks", "totp_identifier")
    op.drop_index(op.f("ix_totp_codes_totp_identifier"), table_name="totp_codes")
    op.drop_index(op.f("ix_totp_codes_expired_at"), table_name="totp_codes")
    op.drop_index(op.f("ix_totp_codes_created_at"), table_name="totp_codes")
    op.drop_table("totp_codes")
    # ### end Alembic commands ###



import asyncio
import time
from collections import defaultdict

import structlog

from skyvern.forge import app
from skyvern.forge.sdk.artifact.models import Artifact, ArtifactType, LogEntityType
from skyvern.forge.sdk.db.id import generate_artifact_id
from skyvern.forge.sdk.models import Step
from skyvern.forge.sdk.schemas.ai_suggestions import AISuggestion
from skyvern.forge.sdk.schemas.task_v2 import TaskV2, Thought
from skyvern.forge.sdk.schemas.workflow_runs import WorkflowRunBlock

LOG = structlog.get_logger(__name__)


class ArtifactManager:
    # task_id -> list of aio_tasks for uploading artifacts
    upload_aiotasks_map: dict[str, list[asyncio.Task[None]]] = defaultdict(list)

    async def _create_artifact(
        self,
        aio_task_primary_key: str,
        artifact_id: str,
        artifact_type: ArtifactType,
        uri: str,
        step_id: str | None = None,
        task_id: str | None = None,
        workflow_run_id: str | None = None,
        workflow_run_block_id: str | None = None,
        thought_id: str | None = None,
        task_v2_id: str | None = None,
        ai_suggestion_id: str | None = None,
        organization_id: str | None = None,
        data: bytes | None = None,
        path: str | None = None,
    ) -> str:
        if data is None and path is None:
            raise ValueError("Either data or path must be provided to create an artifact.")
        if data and path:
            raise ValueError("Both data and path cannot be provided to create an artifact.")
        artifact = await app.DATABASE.create_artifact(
            artifact_id,
            artifact_type,
            uri,
            step_id=step_id,
            task_id=task_id,
            workflow_run_id=workflow_run_id,
            workflow_run_block_id=workflow_run_block_id,
            thought_id=thought_id,
            task_v2_id=task_v2_id,
            organization_id=organization_id,
            ai_suggestion_id=ai_suggestion_id,
        )
        if data:
            # Fire and forget
            aio_task = asyncio.create_task(app.STORAGE.store_artifact(artifact, data))
            self.upload_aiotasks_map[aio_task_primary_key].append(aio_task)
        elif path:
            # Fire and forget
            aio_task = asyncio.create_task(app.STORAGE.store_artifact_from_path(artifact, path))
            self.upload_aiotasks_map[aio_task_primary_key].append(aio_task)

        return artifact_id

    async def create_artifact(
        self,
        step: Step,
        artifact_type: ArtifactType,
        data: bytes | None = None,
        path: str | None = None,
    ) -> str:
        artifact_id = generate_artifact_id()
        uri = app.STORAGE.build_uri(artifact_id, step, artifact_type)
        return await self._create_artifact(
            aio_task_primary_key=step.task_id,
            artifact_id=artifact_id,
            artifact_type=artifact_type,
            uri=uri,
            step_id=step.step_id,
            task_id=step.task_id,
            organization_id=step.organization_id,
            data=data,
            path=path,
        )

    async def create_log_artifact(
        self,
        log_entity_type: LogEntityType,
        log_entity_id: str,
        artifact_type: ArtifactType,
        step_id: str | None = None,
        task_id: str | None = None,
        workflow_run_id: str | None = None,
        workflow_run_block_id: str | None = None,
        organization_id: str | None = None,
        data: bytes | None = None,
        path: str | None = None,
    ) -> str:
        artifact_id = generate_artifact_id()
        uri = app.STORAGE.build_log_uri(log_entity_type, log_entity_id, artifact_type)
        return await self._create_artifact(
            aio_task_primary_key=log_entity_id,
            artifact_id=artifact_id,
            artifact_type=artifact_type,
            uri=uri,
            step_id=step_id,
            task_id=task_id,
            workflow_run_id=workflow_run_id,
            workflow_run_block_id=workflow_run_block_id,
            organization_id=organization_id,
            data=data,
            path=path,
        )

    async def create_thought_artifact(
        self,
        thought: Thought,
        artifact_type: ArtifactType,
        data: bytes | None = None,
        path: str | None = None,
    ) -> str:
        artifact_id = generate_artifact_id()
        uri = app.STORAGE.build_thought_uri(artifact_id, thought, artifact_type)
        return await self._create_artifact(
            aio_task_primary_key=thought.observer_cruise_id,
            artifact_id=artifact_id,
            artifact_type=artifact_type,
            uri=uri,
            thought_id=thought.observer_thought_id,
            task_v2_id=thought.observer_cruise_id,
            organization_id=thought.organization_id,
            data=data,
            path=path,
        )

    async def create_task_v2_artifact(
        self,
        task_v2: TaskV2,
        artifact_type: ArtifactType,
        data: bytes | None = None,
        path: str | None = None,
    ) -> str:
        artifact_id = generate_artifact_id()
        uri = app.STORAGE.build_task_v2_uri(artifact_id, task_v2, artifact_type)
        return await self._create_artifact(
            aio_task_primary_key=task_v2.observer_cruise_id,
            artifact_id=artifact_id,
            artifact_type=artifact_type,
            uri=uri,
            task_v2_id=task_v2.observer_cruise_id,
            organization_id=task_v2.organization_id,
            data=data,
            path=path,
        )

    async def create_workflow_run_block_artifact(
        self,
        workflow_run_block: WorkflowRunBlock,
        artifact_type: ArtifactType,
        data: bytes | None = None,
        path: str | None = None,
    ) -> str:
        artifact_id = generate_artifact_id()
        uri = app.STORAGE.build_workflow_run_block_uri(artifact_id, workflow_run_block, artifact_type)
        return await self._create_artifact(
            aio_task_primary_key=workflow_run_block.workflow_run_block_id,
            artifact_id=artifact_id,
            artifact_type=artifact_type,
            uri=uri,
            workflow_run_block_id=workflow_run_block.workflow_run_block_id,
            workflow_run_id=workflow_run_block.workflow_run_id,
            organization_id=workflow_run_block.organization_id,
            data=data,
            path=path,
        )

    async def create_ai_suggestion_artifact(
        self,
        ai_suggestion: AISuggestion,
        artifact_type: ArtifactType,
        data: bytes | None = None,
        path: str | None = None,
    ) -> str:
        artifact_id = generate_artifact_id()
        uri = app.STORAGE.build_ai_suggestion_uri(artifact_id, ai_suggestion, artifact_type)
        return await self._create_artifact(
            aio_task_primary_key=ai_suggestion.ai_suggestion_id,
            artifact_id=artifact_id,
            artifact_type=artifact_type,
            uri=uri,
            ai_suggestion_id=ai_suggestion.ai_suggestion_id,
            organization_id=ai_suggestion.organization_id,
            data=data,
            path=path,
        )

    async def create_llm_artifact(
        self,
        data: bytes,
        artifact_type: ArtifactType,
        screenshots: list[bytes] | None = None,
        step: Step | None = None,
        thought: Thought | None = None,
        task_v2: TaskV2 | None = None,
        ai_suggestion: AISuggestion | None = None,
    ) -> None:
        if step:
            await self.create_artifact(
                step=step,
                artifact_type=artifact_type,
                data=data,
            )
            for screenshot in screenshots or []:
                await self.create_artifact(
                    step=step,
                    artifact_type=ArtifactType.SCREENSHOT_LLM,
                    data=screenshot,
                )
        elif task_v2:
            await self.create_task_v2_artifact(
                task_v2=task_v2,
                artifact_type=artifact_type,
                data=data,
            )
            for screenshot in screenshots or []:
                await self.create_task_v2_artifact(
                    task_v2=task_v2,
                    artifact_type=ArtifactType.SCREENSHOT_LLM,
                    data=screenshot,
                )
        elif thought:
            await self.create_thought_artifact(
                thought=thought,
                artifact_type=artifact_type,
                data=data,
            )
            for screenshot in screenshots or []:
                await self.create_thought_artifact(
                    thought=thought,
                    artifact_type=ArtifactType.SCREENSHOT_LLM,
                    data=screenshot,
                )
        elif ai_suggestion:
            await self.create_ai_suggestion_artifact(
                ai_suggestion=ai_suggestion,
                artifact_type=artifact_type,
                data=data,
            )
            for screenshot in screenshots or []:
                await self.create_ai_suggestion_artifact(
                    ai_suggestion=ai_suggestion,
                    artifact_type=ArtifactType.SCREENSHOT_LLM,
                    data=screenshot,
                )

    async def update_artifact_data(
        self,
        artifact_id: str | None,
        organization_id: str | None,
        data: bytes,
        primary_key: str = "task_id",
    ) -> None:
        if not artifact_id or not organization_id:
            return None
        artifact = await app.DATABASE.get_artifact_by_id(artifact_id, organization_id)
        if not artifact:
            return
        # Fire and forget
        aio_task = asyncio.create_task(app.STORAGE.store_artifact(artifact, data))

        if not artifact[primary_key]:
            raise ValueError(f"{primary_key} is required to update artifact data.")
        self.upload_aiotasks_map[artifact[primary_key]].append(aio_task)

    async def retrieve_artifact(self, artifact: Artifact) -> bytes | None:
        return await app.STORAGE.retrieve_artifact(artifact)

    async def get_share_link(self, artifact: Artifact) -> str | None:
        return await app.STORAGE.get_share_link(artifact)

    async def get_share_links(self, artifacts: list[Artifact]) -> list[str] | None:
        return await app.STORAGE.get_share_links(artifacts)

    async def wait_for_upload_aiotasks(self, primary_keys: list[str]) -> None:
        try:
            st = time.time()
            async with asyncio.timeout(30):
                await asyncio.gather(
                    *[
                        aio_task
                        for primary_key in primary_keys
                        for aio_task in self.upload_aiotasks_map[primary_key]
                        if not aio_task.done()
                    ]
                )
            LOG.info(
                f"S3 upload aio tasks for primary_keys={primary_keys} completed in {time.time() - st:.2f}s",
                primary_keys=primary_keys,
                duration=time.time() - st,
            )
        except asyncio.TimeoutError:
            LOG.error(
                f"Timeout (30s) while waiting for upload aio tasks for primary_keys={primary_keys}",
                primary_keys=primary_keys,
            )

        for primary_key in primary_keys:
            del self.upload_aiotasks_map[primary_key]



"""change credentials table, add organization_collections table

Revision ID: b4bb0b98912a
Revises: 26c5ed737819
Create Date: 2025-02-20 21:46:23.732969+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "b4bb0b98912a"
down_revision: Union[str, None] = "26c5ed737819"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "organization_bitwarden_collections",
        sa.Column("organization_bitwarden_collection_id", sa.String(), nullable=False),
        sa.Column("organization_id", sa.String(), nullable=False),
        sa.Column("collection_id", sa.String(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("modified_at", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("organization_bitwarden_collection_id"),
    )
    op.create_index(
        op.f("ix_organization_bitwarden_collections_organization_id"),
        "organization_bitwarden_collections",
        ["organization_id"],
        unique=False,
    )
    op.add_column("credentials", sa.Column("item_id", sa.String(), nullable=True))
    op.drop_column("credentials", "website_url")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("credentials", sa.Column("website_url", sa.VARCHAR(), autoincrement=False, nullable=True))
    op.drop_column("credentials", "item_id")
    op.drop_index(
        op.f("ix_organization_bitwarden_collections_organization_id"), table_name="organization_bitwarden_collections"
    )
    op.drop_table("organization_bitwarden_collections")
    # ### end Alembic commands ###



"""artifacts index - workflow run block id index

Revision ID: 172cdfb3e2ee
Revises: d13af1e466fa
Create Date: 2024-12-30 18:16:04.102156+00:00

"""

from typing import Sequence, Union

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "172cdfb3e2ee"
down_revision: Union[str, None] = "d13af1e466fa"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index("org_observer_thought_index", "artifacts", ["organization_id", "observer_thought_id"], unique=False)
    op.create_index("org_wfrb_index", "artifacts", ["organization_id", "workflow_run_block_id"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("org_wfrb_index", table_name="artifacts")
    op.drop_index("org_observer_thought_index", table_name="artifacts")
    # ### end Alembic commands ###



# This file was auto-generated by Fern from our API Definition.

import typing

WorkflowParameterType = typing.Union[
    typing.Literal["string", "integer", "float", "boolean", "json", "file_url", "credential_id"], typing.Any
]



import asyncio
import json
import os
import re
import urllib.parse
from enum import IntEnum, StrEnum
from typing import Tuple

import structlog
import tldextract
from pydantic import BaseModel

from skyvern.config import settings
from skyvern.exceptions import (
    BitwardenAccessDeniedError,
    BitwardenCreateCollectionError,
    BitwardenCreateCreditCardItemError,
    BitwardenCreateLoginItemError,
    BitwardenGetItemError,
    BitwardenListItemsError,
    BitwardenLoginError,
    BitwardenLogoutError,
    BitwardenSecretError,
    BitwardenSyncError,
    BitwardenUnlockError,
)
from skyvern.forge.sdk.api.aws import aws_client
from skyvern.forge.sdk.core.aiohttp_helper import aiohttp_delete, aiohttp_get_json, aiohttp_post
from skyvern.forge.sdk.schemas.credentials import (
    CredentialItem,
    CredentialType,
    CreditCardCredential,
    PasswordCredential,
)

LOG = structlog.get_logger()
BITWARDEN_SERVER_BASE_URL = f"{settings.BITWARDEN_SERVER}:{settings.BITWARDEN_SERVER_PORT or 8002}"


class BitwardenItemType(IntEnum):
    LOGIN = 1
    SECURE_NOTE = 2
    CREDIT_CARD = 3
    IDENTITY = 4


def get_bitwarden_item_type_code(item_type: BitwardenItemType) -> int:
    if item_type == BitwardenItemType.LOGIN:
        return 1
    elif item_type == BitwardenItemType.SECURE_NOTE:
        return 2
    elif item_type == BitwardenItemType.CREDIT_CARD:
        return 3
    elif item_type == BitwardenItemType.IDENTITY:
        return 4


def get_list_response_item_from_bitwarden_item(item: dict) -> CredentialItem:
    if item["type"] == BitwardenItemType.LOGIN:
        login = item["login"]
        totp = BitwardenService.extract_totp_secret(login.get("totp", ""))
        return CredentialItem(
            item_id=item["id"],
            credential=PasswordCredential(
                username=login["username"] or "",
                password=login["password"] or "",
                totp=totp,
            ),
            name=item["name"],
            credential_type=CredentialType.PASSWORD,
        )
    elif item["type"] == BitwardenItemType.CREDIT_CARD:
        card = item["card"]
        return CredentialItem(
            item_id=item["id"],
            credential=CreditCardCredential(
                card_holder_name=card["cardholderName"],
                card_number=card["number"],
                card_exp_month=card["expMonth"],
                card_exp_year=card["expYear"],
                card_cvv=card["code"],
                card_brand=card["brand"],
            ),
            name=item["name"],
            credential_type=CredentialType.CREDIT_CARD,
        )
    else:
        raise BitwardenGetItemError(f"Unsupported item type: {item['type']}")


def is_valid_email(email: str | None) -> bool:
    if not email:
        return False
    pattern = r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$"
    return re.match(pattern, email) is not None


class BitwardenConstants(StrEnum):
    BW_ORGANIZATION_ID = "BW_ORGANIZATION_ID"
    BW_COLLECTION_IDS = "BW_COLLECTION_IDS"

    CLIENT_ID = "BW_CLIENT_ID"
    CLIENT_SECRET = "BW_CLIENT_SECRET"
    MASTER_PASSWORD = "BW_MASTER_PASSWORD"
    URL = "BW_URL"
    BW_COLLECTION_ID = "BW_COLLECTION_ID"
    IDENTITY_KEY = "BW_IDENTITY_KEY"
    BW_ITEM_ID = "BW_ITEM_ID"

    USERNAME = "BW_USERNAME"
    PASSWORD = "BW_PASSWORD"
    TOTP = "BW_TOTP"

    CREDIT_CARD_HOLDER_NAME = "BW_CREDIT_CARD_HOLDER_NAME"
    CREDIT_CARD_NUMBER = "BW_CREDIT_CARD_NUMBER"
    CREDIT_CARD_EXPIRATION_MONTH = "BW_CREDIT_CARD_EXPIRATION_MONTH"
    CREDIT_CARD_EXPIRATION_YEAR = "BW_CREDIT_CARD_EXPIRATION_YEAR"
    CREDIT_CARD_CVV = "BW_CREDIT_CARD_CVV"
    CREDIT_CARD_BRAND = "BW_CREDIT_CARD_BRAND"

    SKYVERN_AUTH_BITWARDEN_ORGANIZATION_ID = "SKYVERN_AUTH_BITWARDEN_ORGANIZATION_ID"
    SKYVERN_AUTH_BITWARDEN_MASTER_PASSWORD = "SKYVERN_AUTH_BITWARDEN_MASTER_PASSWORD"
    SKYVERN_AUTH_BITWARDEN_CLIENT_ID = "SKYVERN_AUTH_BITWARDEN_CLIENT_ID"
    SKYVERN_AUTH_BITWARDEN_CLIENT_SECRET = "SKYVERN_AUTH_BITWARDEN_CLIENT_SECRET"


class BitwardenQueryResult(BaseModel):
    credential: dict[str, str]
    uris: list[str]


class RunCommandResult(BaseModel):
    stdout: str
    stderr: str
    returncode: int


class BitwardenService:
    @staticmethod
    async def run_command(
        command: list[str], additional_env: dict[str, str] | None = None, timeout: int = 60
    ) -> RunCommandResult:
        """
        Run a CLI command with the specified additional environment variables and return the result.
        """
        env = os.environ.copy()  # Copy the current environment
        # Make sure node isn't returning warnings. Warnings are sent through stderr and we raise exceptions on stderr.
        env["NODE_NO_WARNINGS"] = "1"
        if additional_env:
            env.update(additional_env)  # Update with any additional environment variables

        try:
            async with asyncio.timeout(timeout):
                shell_subprocess = await asyncio.create_subprocess_shell(
                    " ".join(command),
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                    env=env,
                )
                stdout, stderr = await shell_subprocess.communicate()
                return RunCommandResult(
                    stdout=stdout.decode(),
                    stderr=stderr.decode(),
                    returncode=shell_subprocess.returncode,
                )
        except asyncio.TimeoutError as e:
            LOG.error(f"Bitwarden command timed out after {timeout} seconds", exc_info=True)
            raise e

    @staticmethod
    def _extract_session_key(unlock_cmd_output: str) -> str | None:
        # Split the text by lines
        lines = unlock_cmd_output.split("\n")

        # Look for the line containing the BW_SESSION
        for line in lines:
            if 'BW_SESSION="' in line:
                # Find the start and end positions of the session key
                start = line.find('BW_SESSION="') + len('BW_SESSION="')
                end = line.rfind('"', start)
                return line[start:end]

        return None

    @staticmethod
    async def get_secret_value_from_url(
        client_id: str,
        client_secret: str,
        master_password: str,
        bw_organization_id: str | None,
        bw_collection_ids: list[str] | None,
        url: str | None = None,
        collection_id: str | None = None,
        item_id: str | None = None,
        max_retries: int = settings.BITWARDEN_MAX_RETRIES,
        timeout: int = settings.BITWARDEN_TIMEOUT_SECONDS,
    ) -> dict[str, str]:
        """
        Get the secret value from the Bitwarden CLI.
        """
        fail_reasons: list[str] = []
        if not bw_organization_id and bw_collection_ids and collection_id not in bw_collection_ids:
            raise BitwardenAccessDeniedError()

        for i in range(max_retries):
            # FIXME: just simply double the timeout for the second try. maybe a better backoff policy when needed
            timeout = (i + 1) * timeout
            try:
                async with asyncio.timeout(timeout):
                    return await BitwardenService._get_secret_value_from_url(
                        client_id=client_id,
                        client_secret=client_secret,
                        master_password=master_password,
                        bw_organization_id=bw_organization_id,
                        bw_collection_ids=bw_collection_ids,
                        url=url,
                        collection_id=collection_id,
                        item_id=item_id,
                        timeout=timeout,
                    )
            except BitwardenAccessDeniedError as e:
                raise e
            except Exception as e:
                LOG.info("Failed to get secret value from Bitwarden", tried_times=i + 1, exc_info=True)
                fail_reasons.append(f"{type(e).__name__}: {str(e)}")
        else:
            raise BitwardenListItemsError(
                f"Bitwarden CLI failed after all retry attempts. Fail reasons: {fail_reasons}"
            )

    @staticmethod
    def extract_totp_secret(totp_value: str) -> str:
        """
        Extract the TOTP secret from either a raw secret or a TOTP URI.

        Args:
            totp_value: Raw TOTP secret or URI (otpauth://totp/...)

        Returns:
            The extracted TOTP secret

        Example:
            >>> BitwardenService.extract_totp_secret("AAAAAABBBBBBB")
            "AAAAAABBBBBBB"
            >>> BitwardenService.extract_totp_secret("otpauth://totp/user@domain.com?secret=AAAAAABBBBBBB")
            "AAAAAABBBBBBB"
        """
        if not totp_value:
            return ""

        # Handle TOTP URI format
        if totp_value.startswith("otpauth://"):
            try:
                # Parse the URI to extract the secret
                query = urllib.parse.urlparse(totp_value).query
                params = dict(urllib.parse.parse_qsl(query))
                return params.get("secret", "")
            except Exception:
                LOG.error(
                    "Failed to parse TOTP URI",
                    totp_value=totp_value,
                    exc_info=True,
                )
                return ""

        return totp_value

    @staticmethod
    async def _get_secret_value_from_url(
        client_id: str,
        client_secret: str,
        master_password: str,
        bw_organization_id: str | None,
        bw_collection_ids: list[str] | None,
        url: str | None = None,
        collection_id: str | None = None,
        item_id: str | None = None,
        timeout: int = 60,
    ) -> dict[str, str]:
        """
        Get the secret value from the Bitwarden CLI.
        """
        try:
            await BitwardenService.login(client_id, client_secret)
            await BitwardenService.sync()
            session_key = await BitwardenService.unlock(master_password)

            if item_id:  # if item_id provided, get single item by item id
                command = ["bw", "get", "item", item_id, "--session", session_key]
                item_result = await BitwardenService.run_command(command)
                if item_result.stderr:
                    raise BitwardenGetItemError(
                        f"Failed to get the bitwarden item {item_id}. Error: {item_result.stderr}"
                    )
                try:
                    item = json.loads(item_result.stdout)
                except json.JSONDecodeError:
                    raise BitwardenGetItemError(f"Failed to parse item JSON for item ID: {item_id}")

                login = item["login"]
                totp = BitwardenService.extract_totp_secret(login.get("totp", ""))

                return {
                    BitwardenConstants.USERNAME: login.get("username", ""),
                    BitwardenConstants.PASSWORD: login.get("password", ""),
                    BitwardenConstants.TOTP: totp,
                }
            elif not url:
                # if item_id is not provided, we need a url to search for items
                raise BitwardenGetItemError("No url or item ID provided")

            # Extract the domain from the URL and search for items in Bitwarden with that domain
            extract_url = tldextract.extract(url)
            domain = extract_url.domain
            list_command = [
                "bw",
                "list",
                "items",
                "--search",
                domain,
                "--session",
                session_key,
            ]
            if bw_organization_id:
                LOG.info(
                    "Organization ID is provided, filtering items by organization ID",
                    bw_organization_id=bw_organization_id,
                )
                list_command.extend(["--organizationid", bw_organization_id])
            elif collection_id:
                LOG.info("Collection ID is provided, filtering items by collection ID", collection_id=collection_id)
                list_command.extend(["--collectionid", collection_id])
            else:
                LOG.error("No collection ID or organization ID provided -- this is required")
                raise BitwardenListItemsError("No collection ID or organization ID provided -- this is required")
            items_result = await BitwardenService.run_command(list_command, timeout=timeout)

            if items_result.stderr and "Event post failed" not in items_result.stderr:
                raise BitwardenListItemsError(items_result.stderr)

            # Parse the items and extract credentials
            try:
                items = json.loads(items_result.stdout)
            except json.JSONDecodeError:
                raise BitwardenListItemsError("Failed to parse items JSON. Output: " + items_result.stdout)

            # Since Bitwarden can't AND multiple filters, we only use organization id in the list command
            # but we still need to filter the items by collection id here
            if bw_organization_id and collection_id:
                filtered_items = []
                for item in items:
                    if "collectionIds" in item and collection_id in item["collectionIds"]:
                        filtered_items.append(item)
                items = filtered_items

            if not items:
                collection_id_str = f" in collection with ID: {collection_id}" if collection_id else ""
                raise BitwardenListItemsError(f"No items found in Bitwarden for URL: {url}{collection_id_str}")

            bitwarden_result: list[BitwardenQueryResult] = []
            for item in items:
                if "login" not in item:
                    continue

                login = item["login"]
                totp = BitwardenService.extract_totp_secret(login.get("totp", ""))

                bitwarden_result.append(
                    BitwardenQueryResult(
                        credential={
                            BitwardenConstants.USERNAME: login.get("username", ""),
                            BitwardenConstants.PASSWORD: login.get("password", ""),
                            BitwardenConstants.TOTP: totp,
                        },
                        uris=[uri.get("uri") for uri in login.get("uris", []) if "uri" in uri],
                    )
                )

            if len(bitwarden_result) == 0:
                return {}

            if len(bitwarden_result) == 1:
                return bitwarden_result[0].credential

            # Choose multiple credentials according to the defined rule,
            # if no cred matches the rule, return the first one.
            # TODO: For now hard code to choose the first matched result
            for single_result in bitwarden_result:
                # check the username is a valid email
                if is_valid_email(single_result.credential.get(BitwardenConstants.USERNAME)):
                    for uri in single_result.uris:
                        # check if the register_domain is the same
                        if extract_url.registered_domain == tldextract.extract(uri).registered_domain:
                            return single_result.credential
            LOG.warning("No credential in Bitwarden matches the rule, returning the first match")
            return bitwarden_result[0].credential
        finally:
            # Step 4: Log out
            await BitwardenService.logout()

    @staticmethod
    async def get_sensitive_information_from_identity(
        client_id: str,
        client_secret: str,
        master_password: str,
        bw_organization_id: str | None,
        bw_collection_ids: list[str] | None,
        collection_id: str,
        identity_key: str,
        identity_fields: list[str],
        remaining_retries: int = settings.BITWARDEN_MAX_RETRIES,
        timeout: int = settings.BITWARDEN_TIMEOUT_SECONDS,
        fail_reasons: list[str] = [],
    ) -> dict[str, str]:
        """
        Get the secret value from the Bitwarden CLI.
        """
        if not bw_organization_id and bw_collection_ids and collection_id not in bw_collection_ids:
            raise BitwardenAccessDeniedError()
        try:
            async with asyncio.timeout(timeout):
                return await BitwardenService._get_sensitive_information_from_identity(
                    client_id=client_id,
                    client_secret=client_secret,
                    master_password=master_password,
                    bw_organization_id=bw_organization_id,
                    bw_collection_ids=bw_collection_ids,
                    collection_id=collection_id,
                    identity_key=identity_key,
                    identity_fields=identity_fields,
                )
        except BitwardenAccessDeniedError as e:
            raise e
        except Exception as e:
            if remaining_retries <= 0:
                raise BitwardenListItemsError(
                    f"Bitwarden CLI failed after all retry attempts. Fail reasons: {fail_reasons}"
                )

            remaining_retries -= 1
            LOG.info("Retrying to get sensitive information from Bitwarden", remaining_retries=remaining_retries)
            return await BitwardenService.get_sensitive_information_from_identity(
                client_id=client_id,
                client_secret=client_secret,
                master_password=master_password,
                bw_organization_id=bw_organization_id,
                bw_collection_ids=bw_collection_ids,
                collection_id=collection_id,
                identity_key=identity_key,
                identity_fields=identity_fields,
                remaining_retries=remaining_retries,
                # Double the timeout for the next retry
                timeout=timeout * 2,
                fail_reasons=fail_reasons + [f"{type(e).__name__}: {str(e)}"],
            )

    @staticmethod
    async def _get_sensitive_information_from_identity(
        client_id: str,
        client_secret: str,
        master_password: str,
        collection_id: str,
        identity_key: str,
        identity_fields: list[str],
        bw_organization_id: str | None,
        bw_collection_ids: list[str] | None,
    ) -> dict[str, str]:
        """
        Get the sensitive information from the Bitwarden CLI.
        """
        try:
            await BitwardenService.login(client_id, client_secret)
            await BitwardenService.sync()
            session_key = await BitwardenService.unlock(master_password)

            if not bw_organization_id and not collection_id:
                raise BitwardenAccessDeniedError()

            # Step 3: Retrieve the items
            list_command = [
                "bw",
                "list",
                "items",
                "--search",
                identity_key,
                "--session",
                session_key,
                "--collectionid",
                collection_id,
            ]
            if bw_organization_id:
                list_command.extend(["--organizationid", bw_organization_id])
            items_result = await BitwardenService.run_command(list_command)

            # Parse the items and extract sensitive information
            try:
                items = json.loads(items_result.stdout)
            except json.JSONDecodeError:
                raise BitwardenListItemsError("Failed to parse items JSON. Output: " + items_result.stdout)
            if not items:
                raise BitwardenListItemsError(
                    f"No items found in Bitwarden for identity key: {identity_key} in collection with ID: {collection_id}"
                )

            # Filter the identity items
            # https://bitwarden.com/help/cli/#create lists the type of the identity items as 4

            # We may want to filter it by type in the future, but for now we just take the first item and check its identity fields
            # identity_items = [item for item in items if item["type"] == 4]

            identity_item = items[0]

            sensitive_information: dict[str, str] = {}
            for field in identity_fields:
                # The identity item may store sensitive information in custom fields or default fields
                # Custom fields are prioritized over default fields
                # TODO (kerem): Make this case insensitive?
                for item in identity_item["fields"]:
                    if item["name"] == field:
                        sensitive_information[field] = item["value"]
                        break

                if (
                    "identity" in identity_item
                    and field in identity_item["identity"]
                    and field not in sensitive_information
                ):
                    sensitive_information[field] = identity_item["identity"][field]

            return sensitive_information

        finally:
            # Step 4: Log out
            await BitwardenService.logout()

    @staticmethod
    async def login(client_id: str, client_secret: str) -> None:
        """
        Log in to the Bitwarden CLI.
        """
        env = {
            "BW_CLIENTID": client_id,
            "BW_CLIENTSECRET": client_secret,
        }
        login_command = ["bw", "login", "--apikey"]
        login_result = await BitwardenService.run_command(login_command, env)

        # Validate the login result
        if login_result.stdout and "You are logged in!" not in login_result.stdout:
            raise BitwardenLoginError(f"Failed to log in. stdout: {login_result.stdout} stderr: {login_result.stderr}")

        if login_result.stderr and "You are already logged in as" not in login_result.stderr:
            raise BitwardenLoginError(f"Failed to log in. stdout: {login_result.stdout} stderr: {login_result.stderr}")

        LOG.info("Bitwarden login successful")

    @staticmethod
    async def unlock(master_password: str) -> str:
        """
        Unlock the Bitwarden CLI.
        """
        env = {
            "BW_PASSWORD": master_password,
        }
        unlock_command = ["bw", "unlock", "--passwordenv", "BW_PASSWORD"]
        unlock_result = await BitwardenService.run_command(unlock_command, env)

        # Validate the unlock result
        if unlock_result.stdout and "Your vault is now unlocked!" not in unlock_result.stdout:
            raise BitwardenUnlockError(
                f"Failed to unlock vault. stdout: {unlock_result.stdout} stderr: {unlock_result.stderr}"
            )

        # Extract session key
        try:
            session_key = BitwardenService._extract_session_key(unlock_result.stdout)
        except Exception as e:
            raise BitwardenUnlockError(f"Unable to extract session key: {str(e)}. stderr: {unlock_result.stderr}")

        if not session_key:
            raise BitwardenUnlockError(f"Session key is empty. stderr: {unlock_result.stderr}")

        return session_key

    @staticmethod
    async def sync() -> None:
        """
        Sync the Bitwarden CLI.
        """
        sync_command = ["bw", "sync"]
        LOG.info("Bitwarden CLI sync started")
        sync_result = await BitwardenService.run_command(sync_command)
        LOG.info("Bitwarden CLI sync completed")
        if sync_result.stderr:
            raise BitwardenSyncError(sync_result.stderr)

    @staticmethod
    async def logout() -> None:
        """
        Log out of the Bitwarden CLI.
        """
        logout_command = ["bw", "logout"]
        logout_result = await BitwardenService.run_command(logout_command)
        if logout_result.stderr and "You are not logged in." not in logout_result.stderr:
            raise BitwardenLogoutError(logout_result.stderr)

    @staticmethod
    async def _get_credit_card_data(
        client_id: str,
        client_secret: str,
        master_password: str,
        bw_organization_id: str | None,
        bw_collection_ids: list[str] | None,
        collection_id: str,
        item_id: str,
    ) -> dict[str, str]:
        """
        Get the credit card data from the Bitwarden CLI.
        """
        try:
            await BitwardenService.login(client_id, client_secret)
            await BitwardenService.sync()
            session_key = await BitwardenService.unlock(master_password)

            # Step 3: Get the item
            get_command = [
                "bw",
                "get",
                "item",
                item_id,
                "--session",
                session_key,
            ]

            # Bitwarden CLI doesn't support filtering by organization ID or collection ID for credit card data so we just raise an error if no collection ID or organization ID is provided
            if not bw_organization_id and not collection_id:
                LOG.error("No collection ID or organization ID provided -- this is required")
                raise BitwardenAccessDeniedError()

            item_result = await BitwardenService.run_command(get_command)

            # Parse the item and extract credit card data
            try:
                item = json.loads(item_result.stdout)
            except json.JSONDecodeError:
                raise BitwardenListItemsError(f"Failed to parse item JSON for item ID: {item_id}")

            if not item:
                raise BitwardenListItemsError(f"No item found in Bitwarden for item ID: {item_id}")

            # Check if the bw_organization_id matches
            if bw_organization_id:
                item_organization_id = item.get("organizationId")
                if item_organization_id != bw_organization_id:
                    raise BitwardenAccessDeniedError()

            if bw_collection_ids:
                item_collection_ids = item.get("collectionIds")
                if item_collection_ids and collection_id not in bw_collection_ids:
                    raise BitwardenAccessDeniedError()

            # Check if the item is a credit card
            # https://bitwarden.com/help/cli/#create lists the type of the credit card items as 3
            if item["type"] != 3:
                raise BitwardenListItemsError(f"Item with ID: {item_id} is not a credit card type")

            credit_card_data = item["card"]

            mapped_credit_card_data: dict[str, str] = {
                BitwardenConstants.CREDIT_CARD_HOLDER_NAME: credit_card_data["cardholderName"],
                BitwardenConstants.CREDIT_CARD_NUMBER: credit_card_data["number"],
                BitwardenConstants.CREDIT_CARD_EXPIRATION_MONTH: credit_card_data["expMonth"],
                BitwardenConstants.CREDIT_CARD_EXPIRATION_YEAR: credit_card_data["expYear"],
                BitwardenConstants.CREDIT_CARD_CVV: credit_card_data["code"],
                BitwardenConstants.CREDIT_CARD_BRAND: credit_card_data["brand"],
            }

            return mapped_credit_card_data
        finally:
            # Step 4: Log out
            await BitwardenService.logout()

    @staticmethod
    async def get_credit_card_data(
        client_id: str,
        client_secret: str,
        master_password: str,
        bw_organization_id: str | None,
        bw_collection_ids: list[str] | None,
        collection_id: str,
        item_id: str,
        remaining_retries: int = settings.BITWARDEN_MAX_RETRIES,
        fail_reasons: list[str] = [],
    ) -> dict[str, str]:
        """
        Get the credit card data from the Bitwarden CLI.
        """
        try:
            async with asyncio.timeout(settings.BITWARDEN_TIMEOUT_SECONDS):
                return await BitwardenService._get_credit_card_data(
                    client_id=client_id,
                    client_secret=client_secret,
                    master_password=master_password,
                    bw_organization_id=bw_organization_id,
                    bw_collection_ids=bw_collection_ids,
                    collection_id=collection_id,
                    item_id=item_id,
                )
        except BitwardenAccessDeniedError as e:
            raise e
        except Exception as e:
            if remaining_retries <= 0:
                raise BitwardenListItemsError(
                    f"Bitwarden CLI failed after all retry attempts. Fail reasons: {fail_reasons}"
                )

            remaining_retries -= 1
            LOG.info("Retrying to get credit card data from Bitwarden", remaining_retries=remaining_retries)
            return await BitwardenService.get_credit_card_data(
                client_id=client_id,
                client_secret=client_secret,
                master_password=master_password,
                bw_organization_id=bw_organization_id,
                bw_collection_ids=bw_collection_ids,
                collection_id=collection_id,
                item_id=item_id,
                remaining_retries=remaining_retries,
                fail_reasons=fail_reasons + [f"{type(e).__name__}: {str(e)}"],
            )

    @staticmethod
    async def _unlock_using_server(master_password: str) -> None:
        status_response = await aiohttp_get_json(f"{BITWARDEN_SERVER_BASE_URL}/status")
        status = status_response["data"]["template"]["status"]
        if status != "unlocked":
            await aiohttp_post(f"{BITWARDEN_SERVER_BASE_URL}/unlock", data={"password": master_password})

    @staticmethod
    async def _get_login_item_by_id_using_server(item_id: str) -> PasswordCredential:
        response = await aiohttp_get_json(f"{BITWARDEN_SERVER_BASE_URL}/object/item/{item_id}")
        if not response or response.get("success") is False:
            raise BitwardenGetItemError(f"Failed to get login item by ID: {item_id}")

        login = response["data"]["login"]
        totp = BitwardenService.extract_totp_secret(login.get("totp", ""))
        if not login:
            raise BitwardenGetItemError(f"Item with ID: {item_id} is not a login item")

        return PasswordCredential(
            username=login["username"] or "",
            password=login["password"] or "",
            totp=totp,
        )

    @staticmethod
    async def _create_login_item_using_server(
        bw_organization_id: str,
        collection_id: str,
        name: str,
        credential: PasswordCredential,
    ) -> str:
        item_template = await aiohttp_get_json(f"{BITWARDEN_SERVER_BASE_URL}/object/template/item")
        login_template = await aiohttp_get_json(f"{BITWARDEN_SERVER_BASE_URL}/object/template/item.login")

        item_template = item_template["data"]["template"]
        login_template = login_template["data"]["template"]

        login_template["username"] = credential.username
        login_template["password"] = credential.password
        login_template["totp"] = credential.totp

        item_template["type"] = get_bitwarden_item_type_code(BitwardenItemType.LOGIN)
        item_template["name"] = name
        item_template["login"] = login_template
        item_template["collectionIds"] = [collection_id]
        item_template["organizationId"] = bw_organization_id

        response = await aiohttp_post(f"{BITWARDEN_SERVER_BASE_URL}/object/item", data=item_template)
        if not response or response.get("success") is False:
            raise BitwardenCreateLoginItemError("Failed to create login item")

        return response["data"]["id"]

    @staticmethod
    async def _create_credit_card_item_using_server(
        bw_organization_id: str,
        collection_id: str,
        name: str,
        credential: CreditCardCredential,
    ) -> str:
        item_template = await aiohttp_get_json(f"{BITWARDEN_SERVER_BASE_URL}/object/template/item")
        credit_card_template = await aiohttp_get_json(f"{BITWARDEN_SERVER_BASE_URL}/object/template/item.card")

        item_template = item_template["data"]["template"]
        credit_card_template = credit_card_template["data"]["template"]

        credit_card_template["cardholderName"] = credential.card_holder_name
        credit_card_template["number"] = credential.card_number
        credit_card_template["expMonth"] = credential.card_exp_month
        credit_card_template["expYear"] = credential.card_exp_year
        credit_card_template["code"] = credential.card_cvv
        credit_card_template["brand"] = credential.card_brand

        item_template["type"] = get_bitwarden_item_type_code(BitwardenItemType.CREDIT_CARD)
        item_template["name"] = name
        item_template["card"] = credit_card_template
        item_template["collectionIds"] = [collection_id]
        item_template["organizationId"] = bw_organization_id

        response = await aiohttp_post(f"{BITWARDEN_SERVER_BASE_URL}/object/item", data=item_template)
        if not response or response.get("success") is False:
            raise BitwardenCreateCreditCardItemError("Failed to create credit card item")

        return response["data"]["id"]

    @staticmethod
    async def create_credential_item(
        collection_id: str,
        name: str,
        credential: PasswordCredential | CreditCardCredential,
    ) -> str:
        try:
            master_password, bw_organization_id, _, _ = await BitwardenService._get_skyvern_auth_secrets()

            await BitwardenService._unlock_using_server(master_password)
            if isinstance(credential, PasswordCredential):
                return await BitwardenService._create_login_item_using_server(
                    bw_organization_id=bw_organization_id,
                    collection_id=collection_id,
                    name=name,
                    credential=credential,
                )
            else:
                return await BitwardenService._create_credit_card_item_using_server(
                    bw_organization_id=bw_organization_id,
                    collection_id=collection_id,
                    name=name,
                    credential=credential,
                )
        except Exception as e:
            raise e

    @staticmethod
    async def _get_skyvern_auth_master_password() -> str:
        master_password = settings.SKYVERN_AUTH_BITWARDEN_MASTER_PASSWORD
        if not master_password:
            master_password = await aws_client.get_secret(BitwardenConstants.SKYVERN_AUTH_BITWARDEN_MASTER_PASSWORD)
        if not master_password:
            raise BitwardenSecretError("Skyvern auth master password is not set")
        return master_password

    @staticmethod
    async def _get_skyvern_auth_organization_id() -> str:
        bw_organization_id = settings.SKYVERN_AUTH_BITWARDEN_ORGANIZATION_ID
        if not bw_organization_id:
            bw_organization_id = await aws_client.get_secret(BitwardenConstants.SKYVERN_AUTH_BITWARDEN_ORGANIZATION_ID)
        if not bw_organization_id:
            raise BitwardenSecretError("Skyvern auth organization ID is not set")
        return bw_organization_id

    @staticmethod
    async def _get_skyvern_auth_client_id() -> str:
        client_id = settings.SKYVERN_AUTH_BITWARDEN_CLIENT_ID
        if not client_id:
            client_id = await aws_client.get_secret(BitwardenConstants.SKYVERN_AUTH_BITWARDEN_CLIENT_ID)
        if not client_id:
            raise BitwardenSecretError("Skyvern auth client ID is not set")
        return client_id

    @staticmethod
    async def _get_skyvern_auth_client_secret() -> str:
        client_secret = settings.SKYVERN_AUTH_BITWARDEN_CLIENT_SECRET
        if not client_secret:
            client_secret = await aws_client.get_secret(BitwardenConstants.SKYVERN_AUTH_BITWARDEN_CLIENT_SECRET)
        if not client_secret:
            raise BitwardenSecretError("Skyvern auth client secret is not set")
        return client_secret

    @staticmethod
    async def create_collection(
        name: str,
    ) -> str:
        """
        Create a collection in Bitwarden and return the collection ID.
        """
        try:
            master_password, bw_organization_id, _, _ = await BitwardenService._get_skyvern_auth_secrets()

            await BitwardenService._unlock_using_server(master_password)
            return await BitwardenService._create_collection_using_server(bw_organization_id, name)

        except Exception as e:
            raise e

    @staticmethod
    async def _create_collection_using_server(bw_organization_id: str, name: str) -> str:
        collection_template_response = await aiohttp_get_json(f"{BITWARDEN_SERVER_BASE_URL}/object/template/collection")
        collection_template = collection_template_response["data"]["template"]

        collection_template["name"] = name
        collection_template["organizationId"] = bw_organization_id

        response = await aiohttp_post(
            f"{BITWARDEN_SERVER_BASE_URL}/object/org-collection?organizationId={bw_organization_id}",
            data=collection_template,
        )
        if not response or response.get("success") is False:
            raise BitwardenCreateCollectionError("Failed to create collection")

        return response["data"]["id"]

    @staticmethod
    async def _get_skyvern_auth_secrets() -> Tuple[str, str, str, str]:
        master_password, bw_organization_id, client_id, client_secret = await asyncio.gather(
            BitwardenService._get_skyvern_auth_master_password(),
            BitwardenService._get_skyvern_auth_organization_id(),
            BitwardenService._get_skyvern_auth_client_id(),
            BitwardenService._get_skyvern_auth_client_secret(),
        )
        return master_password, bw_organization_id, client_id, client_secret

    @staticmethod
    async def get_items_by_item_ids(
        item_ids: list[str],
    ) -> list[CredentialItem]:
        try:
            master_password, _, _, _ = await BitwardenService._get_skyvern_auth_secrets()
            await BitwardenService._unlock_using_server(master_password)
            return await BitwardenService._get_items_by_item_ids_using_server(item_ids)
        except Exception as e:
            raise e

    @staticmethod
    async def _get_items_by_item_ids_using_server(item_ids: list[str]) -> list[CredentialItem]:
        responses = await asyncio.gather(
            *[aiohttp_get_json(f"{BITWARDEN_SERVER_BASE_URL}/object/item/{item_id}") for item_id in item_ids]
        )
        if not responses or any(response.get("success") is False for response in responses):
            raise BitwardenGetItemError("Failed to get collection items")

        return [get_list_response_item_from_bitwarden_item(response["data"]) for response in responses]

    @staticmethod
    async def get_collection_items(
        collection_id: str,
    ) -> list[CredentialItem]:
        try:
            master_password, _, _, _ = await BitwardenService._get_skyvern_auth_secrets()
            await BitwardenService._unlock_using_server(master_password)
            return await BitwardenService._get_collection_items_using_server(collection_id)
        except Exception as e:
            raise e

    @staticmethod
    async def _get_collection_items_using_server(collection_id: str) -> list[CredentialItem]:
        response = await aiohttp_get_json(f"{BITWARDEN_SERVER_BASE_URL}/list/object/items?collectionId={collection_id}")
        if not response or response.get("success") is False:
            raise BitwardenGetItemError("Failed to get collection items")

        items = response["data"]["data"]
        items = map(lambda item: get_list_response_item_from_bitwarden_item(item), items)
        return list(items)

    @staticmethod
    async def get_credential_item(
        item_id: str,
    ) -> CredentialItem:
        try:
            master_password, _, _, _ = await BitwardenService._get_skyvern_auth_secrets()
            await BitwardenService._unlock_using_server(master_password)
            return await BitwardenService._get_credential_item_by_id_using_server(item_id)
        except Exception as e:
            raise e

    @staticmethod
    async def _get_credential_item_by_id_using_server(item_id: str) -> CredentialItem:
        response = await aiohttp_get_json(f"{BITWARDEN_SERVER_BASE_URL}/object/item/{item_id}")
        if not response or response.get("success") is False:
            raise BitwardenGetItemError(f"Failed to get credential item by ID: {item_id}")

        if response["data"]["type"] == BitwardenItemType.LOGIN:
            login_item = response["data"]["login"]
            name = response["data"]["name"]
            return CredentialItem(
                item_id=item_id,
                credential_type=CredentialType.PASSWORD,
                name=name,
                credential=PasswordCredential(
                    username=login_item["username"] or "",
                    password=login_item["password"] or "",
                    totp=login_item["totp"],
                ),
            )
        elif response["data"]["type"] == BitwardenItemType.CREDIT_CARD:
            credit_card_item = response["data"]["card"]
            name = response["data"]["name"]
            return CredentialItem(
                item_id=item_id,
                credential_type=CredentialType.CREDIT_CARD,
                name=name,
                credential=CreditCardCredential(
                    card_holder_name=credit_card_item["cardholderName"],
                    card_number=credit_card_item["number"],
                    card_exp_month=credit_card_item["expMonth"],
                    card_exp_year=credit_card_item["expYear"],
                    card_cvv=credit_card_item["code"],
                    card_brand=credit_card_item["brand"],
                ),
            )
        else:
            raise BitwardenGetItemError(f"Unsupported item type: {response['data']['type']}")

    @staticmethod
    async def delete_credential_item(
        item_id: str,
    ) -> None:
        try:
            master_password, _, _, _ = await BitwardenService._get_skyvern_auth_secrets()
            await BitwardenService._unlock_using_server(master_password)
            await BitwardenService._delete_credential_item_using_server(item_id)
        except Exception as e:
            raise e

    @staticmethod
    async def _delete_credential_item_using_server(item_id: str) -> None:
        await aiohttp_delete(f"{BITWARDEN_SERVER_BASE_URL}/object/item/{item_id}")



from pydantic_settings import BaseSettings, SettingsConfigDict

from skyvern import constants
from skyvern.constants import SKYVERN_DIR


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_file=(".env", ".env.staging", ".env.prod"), extra="ignore")

    ADDITIONAL_MODULES: list[str] = []

    BROWSER_TYPE: str = "chromium-headful"
    BROWSER_REMOTE_DEBUGGING_URL: str = "http://127.0.0.1:9222"
    CHROME_EXECUTABLE_PATH: str | None = None
    MAX_SCRAPING_RETRIES: int = 0
    VIDEO_PATH: str | None = "./video"
    HAR_PATH: str | None = "./har"
    LOG_PATH: str = "./log"
    TEMP_PATH: str = "./temp"
    BROWSER_ACTION_TIMEOUT_MS: int = 5000
    BROWSER_SCREENSHOT_TIMEOUT_MS: int = 20000
    BROWSER_LOADING_TIMEOUT_MS: int = 90000
    OPTION_LOADING_TIMEOUT_MS: int = 600000
    MAX_STEPS_PER_RUN: int = 10
    MAX_STEPS_PER_TASK_V2: int = 25
    MAX_ITERATIONS_PER_TASK_V2: int = 10
    MAX_NUM_SCREENSHOTS: int = 10
    # Ratio should be between 0 and 1.
    # If the task has been running for more steps than this ratio of the max steps per run, then we'll log a warning.
    LONG_RUNNING_TASK_WARNING_RATIO: float = 0.95
    MAX_RETRIES_PER_STEP: int = 5
    DEBUG_MODE: bool = False
    DATABASE_STRING: str = "postgresql+psycopg://skyvern@localhost/skyvern"
    DATABASE_STATEMENT_TIMEOUT_MS: int = 60000
    DISABLE_CONNECTION_POOL: bool = False
    PROMPT_ACTION_HISTORY_WINDOW: int = 1
    TASK_RESPONSE_ACTION_SCREENSHOT_COUNT: int = 3

    ENV: str = "local"
    EXECUTE_ALL_STEPS: bool = True
    JSON_LOGGING: bool = False
    LOG_LEVEL: str = "INFO"
    PORT: int = 8000
    ALLOWED_ORIGINS: list[str] = ["*"]
    BLOCKED_HOSTS: list[str] = ["localhost"]
    ALLOWED_HOSTS: list[str] = []

    # Format: "http://<username>:<password>@host:port, http://<username>:<password>@host:port, ...."
    HOSTED_PROXY_POOL: str = ""
    ENABLE_PROXY: bool = False

    # Secret key for JWT. Please generate your own secret key in production
    SECRET_KEY: str = "PLACEHOLDER"
    # Algorithm used to sign the JWT
    SIGNATURE_ALGORITHM: str = "HS256"
    ACCESS_TOKEN_EXPIRE_MINUTES: int = 60 * 24 * 7  # one week

    # Artifact storage settings
    ARTIFACT_STORAGE_PATH: str = f"{SKYVERN_DIR}/artifacts"
    GENERATE_PRESIGNED_URLS: bool = False
    AWS_S3_BUCKET_ARTIFACTS: str = "skyvern-artifacts"
    AWS_S3_BUCKET_SCREENSHOTS: str = "skyvern-screenshots"
    AWS_S3_BUCKET_BROWSER_SESSIONS: str = "skyvern-browser-sessions"

    # Supported storage types: local, s3
    SKYVERN_STORAGE_TYPE: str = "local"

    # S3 bucket settings
    AWS_REGION: str = "us-east-1"
    AWS_S3_BUCKET_UPLOADS: str = "skyvern-uploads"
    MAX_UPLOAD_FILE_SIZE: int = 10 * 1024 * 1024  # 10 MB
    PRESIGNED_URL_EXPIRATION: int = 60 * 60 * 24  # 24 hours

    SKYVERN_TELEMETRY: bool = True
    ANALYTICS_ID: str = "anonymous"

    # browser settings
    BROWSER_LOCALE: str = "en-US"
    BROWSER_TIMEZONE: str = "America/New_York"
    BROWSER_WIDTH: int = 1920
    BROWSER_HEIGHT: int = 1080

    # Add extension folders name here to load extension in your browser
    EXTENSIONS_BASE_PATH: str = "./extensions"
    EXTENSIONS: list[str] = []

    # Workflow constant parameters
    WORKFLOW_DOWNLOAD_DIRECTORY_PARAMETER_KEY: str = "SKYVERN_DOWNLOAD_DIRECTORY"
    WORKFLOW_WAIT_BLOCK_MAX_SEC: int = 30 * 60

    # Saved browser session settings
    BROWSER_SESSION_BASE_PATH: str = f"{constants.REPO_ROOT_DIR}/browser_sessions"

    #####################
    # Bitwarden Configs #
    #####################
    BITWARDEN_TIMEOUT_SECONDS: int = 60
    BITWARDEN_MAX_RETRIES: int = 2

    # task generation settings
    PROMPT_CACHE_WINDOW_HOURS: int = 24

    #####################
    # LLM Configuration #
    #####################
    # ACTIVE LLM PROVIDER
    LLM_KEY: str = "OPENAI_GPT4O"  # This is the model name
    LLM_API_KEY: str | None = None  # API key for the model
    SECONDARY_LLM_KEY: str | None = None
    SELECT_AGENT_LLM_KEY: str | None = None
    SINGLE_CLICK_AGENT_LLM_KEY: str | None = None
    PROMPT_BLOCK_LLM_KEY: str | None = None
    # COMMON
    LLM_CONFIG_TIMEOUT: int = 300
    LLM_CONFIG_MAX_TOKENS: int = 4096
    LLM_CONFIG_TEMPERATURE: float = 0
    LLM_CONFIG_SUPPORT_VISION: bool = True  # Whether the model supports vision
    LLM_CONFIG_ADD_ASSISTANT_PREFIX: bool = False  # Whether to add assistant prefix
    # LLM PROVIDER SPECIFIC
    ENABLE_OPENAI: bool = False
    ENABLE_ANTHROPIC: bool = False
    ENABLE_BEDROCK_ANTHROPIC: bool = False
    ENABLE_AZURE: bool = False
    ENABLE_AZURE_GPT4O_MINI: bool = False
    ENABLE_AZURE_O3_MINI: bool = False
    ENABLE_BEDROCK: bool = False
    ENABLE_GEMINI: bool = False
    ENABLE_VERTEX_AI: bool = False
    ENABLE_AZURE_CUA: bool = False
    ENABLE_OPENAI_COMPATIBLE: bool = False
    # OPENAI
    OPENAI_API_KEY: str | None = None
    # ANTHROPIC
    ANTHROPIC_API_KEY: str | None = None
    ANTHROPIC_CUA_LLM_KEY: str = "ANTHROPIC_CLAUDE3.7_SONNET"

    # OPENAI COMPATIBLE
    OPENAI_COMPATIBLE_MODEL_NAME: str | None = None
    OPENAI_COMPATIBLE_API_KEY: str | None = None
    OPENAI_COMPATIBLE_API_BASE: str | None = None
    OPENAI_COMPATIBLE_API_VERSION: str | None = None
    OPENAI_COMPATIBLE_MAX_TOKENS: int | None = None
    OPENAI_COMPATIBLE_TEMPERATURE: float | None = None
    OPENAI_COMPATIBLE_SUPPORTS_VISION: bool = False
    OPENAI_COMPATIBLE_ADD_ASSISTANT_PREFIX: bool = False
    OPENAI_COMPATIBLE_MODEL_KEY: str = "OPENAI_COMPATIBLE"
    OPENAI_COMPATIBLE_REASONING_EFFORT: str | None = None

    # AZURE
    AZURE_DEPLOYMENT: str | None = None
    AZURE_API_KEY: str | None = None
    AZURE_API_BASE: str | None = None
    AZURE_API_VERSION: str | None = None
    AZURE_CUA_API_KEY: str | None = None
    AZURE_CUA_ENDPOINT: str | None = None
    AZURE_CUA_DEPLOYMENT: str | None = "computer-use-preview"
    AZURE_CUA_API_VERSION: str | None = "2025-03-01-preview"

    # AZURE GPT-4o mini
    AZURE_GPT4O_MINI_DEPLOYMENT: str | None = None
    AZURE_GPT4O_MINI_API_KEY: str | None = None
    AZURE_GPT4O_MINI_API_BASE: str | None = None
    AZURE_GPT4O_MINI_API_VERSION: str | None = None

    # AZURE o3 mini
    AZURE_O3_MINI_DEPLOYMENT: str | None = None
    AZURE_O3_MINI_API_KEY: str | None = None
    AZURE_O3_MINI_API_BASE: str | None = None
    AZURE_O3_MINI_API_VERSION: str | None = None

    # AZURE gpt-4.1
    ENABLE_AZURE_GPT4_1: bool = False
    AZURE_GPT4_1_DEPLOYMENT: str = "gpt-4.1"
    AZURE_GPT4_1_API_KEY: str | None = None
    AZURE_GPT4_1_API_BASE: str | None = None
    AZURE_GPT4_1_API_VERSION: str = "2025-01-01-preview"

    # AZURE gpt-4.1 mini
    ENABLE_AZURE_GPT4_1_MINI: bool = False
    AZURE_GPT4_1_MINI_DEPLOYMENT: str = "gpt-4.1-mini"
    AZURE_GPT4_1_MINI_API_KEY: str | None = None
    AZURE_GPT4_1_MINI_API_BASE: str | None = None
    AZURE_GPT4_1_MINI_API_VERSION: str = "2025-01-01-preview"

    # AZURE gpt-4.1 nano
    ENABLE_AZURE_GPT4_1_NANO: bool = False
    AZURE_GPT4_1_NANO_DEPLOYMENT: str = "gpt-4.1-nano"
    AZURE_GPT4_1_NANO_API_KEY: str | None = None
    AZURE_GPT4_1_NANO_API_BASE: str | None = None
    AZURE_GPT4_1_NANO_API_VERSION: str = "2025-01-01-preview"

    # AZURE o4-mini
    ENABLE_AZURE_O4_MINI: bool = False
    AZURE_O4_MINI_DEPLOYMENT: str = "o4-mini"
    AZURE_O4_MINI_API_KEY: str | None = None
    AZURE_O4_MINI_API_BASE: str | None = None
    AZURE_O4_MINI_API_VERSION: str = "2025-01-01-preview"

    # AZURE o3
    ENABLE_AZURE_O3: bool = False
    AZURE_O3_DEPLOYMENT: str = "o3"
    AZURE_O3_API_KEY: str | None = None
    AZURE_O3_API_BASE: str | None = None
    AZURE_O3_API_VERSION: str = "2025-01-01-preview"

    # GEMINI
    GEMINI_API_KEY: str | None = None

    # VERTEX_AI
    VERTEX_CREDENTIALS: str | None = None

    # NOVITA AI
    ENABLE_NOVITA: bool = False
    NOVITA_API_KEY: str | None = None
    NOVITA_API_VERSION: str = "v3"

    # TOTP Settings
    TOTP_LIFESPAN_MINUTES: int = 10
    VERIFICATION_CODE_INITIAL_WAIT_TIME_SECS: int = 40
    VERIFICATION_CODE_POLLING_TIMEOUT_MINS: int = 15

    # Bitwarden Settings
    BITWARDEN_CLIENT_ID: str | None = None
    BITWARDEN_CLIENT_SECRET: str | None = None
    BITWARDEN_MASTER_PASSWORD: str | None = None

    # Skyvern Auth Bitwarden Settings
    SKYVERN_AUTH_BITWARDEN_CLIENT_ID: str | None = None
    SKYVERN_AUTH_BITWARDEN_CLIENT_SECRET: str | None = None
    SKYVERN_AUTH_BITWARDEN_MASTER_PASSWORD: str | None = None
    SKYVERN_AUTH_BITWARDEN_ORGANIZATION_ID: str | None = None

    BITWARDEN_SERVER: str = "http://localhost"
    BITWARDEN_SERVER_PORT: int = 8002

    SVG_MAX_LENGTH: int = 100000

    ENABLE_LOG_ARTIFACTS: bool = False
    ENABLE_CODE_BLOCK: bool = False

    # SkyvernClient Settings
    SKYVERN_BASE_URL: str = "https://api.skyvern.com"
    SKYVERN_API_KEY: str = "PLACEHOLDER"

    def is_cloud_environment(self) -> bool:
        """
        :return: True if env is not local, else False
        """
        return self.ENV != "local"

    def execute_all_steps(self) -> bool:
        """
        This provides the functionality to execute steps one by one through the Streamlit UI.
        ***Value is always True if ENV is not local.***

        :return: True if env is not local, else the value of EXECUTE_ALL_STEPS
        """
        if self.is_cloud_environment():
            return True
        else:
            return self.EXECUTE_ALL_STEPS


settings = Settings()



from skyvern.config import Settings
from skyvern.config import settings as base_settings


class SettingsManager:
    __instance: Settings = base_settings

    @staticmethod
    def get_settings() -> Settings:
        return SettingsManager.__instance

    @staticmethod
    def set_settings(settings: Settings) -> None:
        SettingsManager.__instance = settings



from skyvern.forge.sdk.cache.base import BaseCache
from skyvern.forge.sdk.cache.local import LocalCache


class CacheFactory:
    __cache: BaseCache = LocalCache()

    @staticmethod
    def set_cache(cache: BaseCache) -> None:
        CacheFactory.__cache = cache

    @staticmethod
    def get_cache() -> BaseCache:
        return CacheFactory.__cache
